import asyncio
import base64
import json
from typing import Callable, Awaitable, Literal, Dict
import numpy as np
import websockets
from starlette.websockets import WebSocket
from websockets.protocol import State
import samplerate
from gallama.data_classes import UsageResponse, ChatCompletionResponse, WSInterTTS, ChoiceDeltaToolCall
from gallama.data_classes.realtime_client_proto import ResponseCancel,\
    ResponseCreate, SessionConfig
from gallama.data_classes.realtime_server_proto import *
from gallama.data_classes.internal_ws import WSInterCancel
from gallama.realtime.websocket_client import WebSocketClient

from ..dependencies_server import get_server_logger

logger = get_server_logger()

class Response:
    """ this class managing the current response to client"""
    def __init__(
        self,
        session: "WebSocketSession",
        ws_client: WebSocket,
        ws_stt: WebSocketClient,
        ws_llm: WebSocketClient,
        ws_tts: WebSocketClient,
        response_id: str,
        event_id_generator: Callable[[], Awaitable[str]]
    ):
        self.session = session
        self.ws_client = ws_client      # this ws is end user
        self.ws_stt = ws_stt
        self.ws_llm = ws_llm
        self.ws_tts = ws_tts

        self.tts_queue = asyncio.Queue()

        # self.current_session_config: SessionConfig = None
        self.usage: UsageResponseRealTime = UsageResponseRealTime()

        self.response_id = response_id
        self.event_id_generator = event_id_generator

        self.item_id: str = None
        self.previous_item_id: str = None

        self.audio: np.ndarray = np.array([], dtype=np.float32)
        self.audio_done: bool = False

        self.text: str = ""
        self.text_done: bool = False

        self.transcription: str = ""
        self.transcription_done: bool = False

        self.function_calling_arguments = None
        self.function_calling_name = None
        self.function_calling_tool_id = None


        # some tracker
        self.content_part_done_status: bool = False
        self.output_item_done_status: bool = False
        self.response_done_status: bool = False
        self.conversation_item_created_status: bool = False     # track the initial conversation created

        # Add cancellation control
        self.cancel_event = asyncio.Event()


    async def cancel(self, cancel_event: ResponseCancel):
        """Cancel the current response while preserving partial results"""
        try:
            self.cancel_event.set()

            # send event to llm and tts to stop generation
            if self.ws_llm and self.ws_llm.connection:
                await self.ws_llm.send_pydantic_message(WSInterCancel(type="common.cancel"))

            # Cancel TTS generation
            if self.ws_tts and self.ws_tts.connection:
                await self.ws_tts.send_pydantic_message(WSInterCancel(type="common.cancel"))
        except Exception as e:
            logger.error(f"Error during response cancellation: {str(e)}")
            raise

    async def content_part_done(self, generation_type: Literal["text", "audio", "transcription", "function_call"]):
        logger.info(f"content part done is called with type: {generation_type}")
        logger.debug(f"transcription: {self.transcription_done}")
        logger.debug(f"function_call: {self.function_calling_name}")
        logger.debug(f"audio: {self.audio_done}")
        logger.debug(f"text: {self.text_done}")

        done_check: bool = False
        content_part: Dict = {}

        # for text type, just the text finish is enough
        allowed_modalities = self.session.config.modalities
        if generation_type == "function_call":
            # currently no content part done response for function call
            done_check = True
            pass

        elif generation_type == "text" and "text" in allowed_modalities and "audio" not in allowed_modalities:
            if self.text_done:
                done_check = True
                content_part = {
                    "type": "text",
                    "text": self.text
                }
            else:
                raise Exception("text generation not finished but content part is marked as done")

        # for audio type, we need both transcript and audio to finish
        if generation_type in ["audio", "transcription"] and "audio" in allowed_modalities:
            if generation_type=="transcription" and not self.transcription_done:
                raise Exception("Transcription generation not finished but content part is marked as done")

            if generation_type=="audio" and not self.audio_done:
                raise Exception("Audio generation not finished but content part is marked as done")

            if self.transcription_done and self.audio_done:
                done_check = True
                content_part = {
                    "type": "audio",
                    #"audio": self.audio,       # this is not included in data sent to client, despite what said in documentation
                    "transcript": self.transcription
                }

        # if done_check achieved meaning text finished or audio & transcription finished
        if done_check:
            # no content part done for function call
            if not generation_type=="function_call":
                await self.ws_client.send_json(ResponseContentPartDoneEvent(**{
                    "event_id": await self.event_id_generator(),
                    "type": "response.content_part.done",
                    "response_id": self.response_id,
                    "item_id": self.item_id,
                    "output_index": 0,
                    "content_index": 0,
                    "part": content_part
                }).model_dump(exclude_unset=True))  # if not, None will be set as audio to front end

            # mark status of content_part as done
            self.content_part_done_status = True
            logger.info(f"Content part done")

            # trigger response output item done
            await self.response_output_item_done(mode=generation_type)

    async def response_output_item_done(self, mode:Literal["text", "audio", "transcription", "function_call"]):

        if mode=="audio" or mode=="transcription":
            _type = "audio"
        elif mode=="text":
            _type = "text"
        elif mode=="function_call":
            _type = "function_call"
        else:
            raise Exception("Unknown mode in response_output_item_done()")

        item = None

        if mode != "function_call":
            item = ConversationItemMessageServer(
                id=self.item_id,
                object="realtime.item",
                type="message",
                status="completed",
                role="assistant",
                content=[MessageContentServer(
                    type=_type,
                    text=self.text if _type=="text" else None,
                    transcript=self.transcription if _type=="audio" else None,
                )]
            )
        else:
            item = ConversationItemFunctionCallServer(
                id=self.item_id,
                object="realtime.item",
                status="completed",
                type="function_call",
                call_id=self.function_calling_tool_id,
                name=self.function_calling_name,
                arguments=self.function_calling_arguments
            )

        # send user response output_item done
        await self.ws_client.send_json(ResponseOutput_ItemDone(**{
            "event_id": await self.event_id_generator(),
            "type": "response.output_item.done",
            "response_id": self.response_id,
            "output_index": 0,
            "item": item.model_dump(exclude_unset=True)
        }).model_dump())

        self.output_item_done_status = True

        # trigger Response done
        await self.response_done()


    async def response_done(self):
        try:
            content_part_status = "completed" if not self.cancel_event.is_set() else "incomplete"
            server_response_status = "completed" if not self.cancel_event.is_set() else "cancelled"

            # content part and response item must be done first
            if not self.response_done_status and self.content_part_done_status and self.output_item_done_status:
                _output = None
                allowed_modalities = self.session.config.modalities

                if self.function_calling_name:  # function calling check first
                    _output = parse_conversation_item({
                        "id": self.item_id,
                        "object": "realtime.item",
                        "type": "function_call",
                        "status": content_part_status,
                        "name": self.function_calling_name,
                        "call_id": self.function_calling_tool_id,
                        "arguments": self.function_calling_arguments,
                    })
                elif "audio" in allowed_modalities:
                    _output = parse_conversation_item({
                        "id": self.item_id,
                        "object": "realtime.item",
                        "type": "message",
                        "status": content_part_status,
                        "role": "assistant",
                        "content": [
                            {
                                "type": "audio",
                                # "audio": self.audio,
                                "audio": None,      # different to documentation, audio is not resent, also this is in bytes, not base64
                                "transcript": self.transcription
                            }
                        ]
                    })
                elif "text" in allowed_modalities and "audio" not in allowed_modalities:
                    # send response done
                    _output = parse_conversation_item({
                        "id": self.item_id,
                        "object": "realtime.item",
                        "type": "message",
                        "status": content_part_status,
                        "role": "assistant",
                        "content": [
                            {
                                "type": "text",
                                "text": self.text
                            }
                        ]
                    })
                else:
                    raise Exception("Unknown modalities type in response_done()")

                # send user status
                if _output:
                    _response = ServerResponse(**{
                        "id": self.response_id,
                        "object": "realtime.response",
                        "status": server_response_status,
                        "usage": self.usage.model_dump(),
                        "output": [_output.model_dump()]        # it must be an array
                    })

                    await self.ws_client.send_json(ResponseDone(**{
                        "event_id": await self.event_id_generator(),
                        "type": "response.done",
                        "response_id": self.response_id,
                        "item_id": self.item_id,
                        "response": _response.model_dump()
                    }).model_dump())
                else:
                    raise Exception("Invalid Response Type")

                # update response_done tracker
                self.response_done_status = True
            elif self.response_done_status:
                raise Exception("Response already sent error in response_done()")
            elif not self.content_part_done_status:
                raise Exception("Content part done must be sent before response done, in response_done()")

        except Exception as e:
            logger.error(f"Error in response_done(): {e}")


    async def response_initialize(self):
        """ send empty event to client to signal response is created """
        await self.ws_client.send_json(ResponseCreated(**{
            "event_id": await self.event_id_generator(),
            "type": "response.created",
            "response": {
                "id": self.response_id,
                "status": "in_progress",
            }
        }).model_dump())

        # register the item id
        self.item_id, self.previous_item_id = await self.session.queues.next_item(return_current=True)


    async def send_conversation_item_created(
        self,
        mode: Literal["text", "transcription", "audio", "function_call"],
        function_name: str = None,  # for function calling
        call_id: str = None,        # for function calling
    ):
        if mode != "function_call":
            initial_item_created_object = ConversationItemCreated(**{
                "event_id": await self.event_id_generator(),
                "type": "conversation.item.created",
                "previous_item_id": self.previous_item_id,
                "item": {
                    "id": self.item_id,
                    "object": "realtime.item",
                    "type": "message",
                    "status": "in_progress",
                    "role": "assistant",
                    "content": []
                }
            })
            await self.session.queues.update_conversation_item_ordered_dict(
                ws_client=self.ws_client,
                ws_llm=self.ws_llm,
                item=initial_item_created_object.item,
            )

            # send initial output_item created
            await self.ws_client.send_json(ResponseOutput_ItemAdded(**{
                "event_id": await self.event_id_generator(),
                "response_id": self.response_id,
                "output_index": 0,
                "item": initial_item_created_object.item
            }).model_dump())

        elif mode == "function_call":   # function call
            initial_item_created_object = ConversationItemCreated(**{
                "event_id": await self.event_id_generator(),
                "type": "conversation.item.created",
                "previous_item_id": self.previous_item_id,
                "item": {
                    "id": self.item_id,
                    "object": "realtime.item",
                    "type": "function_call",
                    "status": "in_progress",
                    "name": function_name,
                    "call_id": call_id,
                }
            })
            await self.session.queues.update_conversation_item_ordered_dict(
                ws_client=self.ws_client,
                ws_llm=self.ws_llm,
                item=initial_item_created_object.item
            )

            # send initial output_item created
            await self.ws_client.send_json(ResponseOutput_ItemAdded(**{
                "event_id": await self.event_id_generator(),
                "response_id": self.response_id,
                "output_index": 0,
                "item": initial_item_created_object.item
            }).model_dump())
        else:
            logger.error(f"Unknown mode in send_conversation_item_created() of mode: {mode}")


        # send content_part added for text to let front end know that there is text coming
        if mode == "audio" or mode == "transcription":
            # Send initial audio content part added event
            await self.ws_client.send_json(ResponseContentPartAddedEvent(**{
                "event_id": await self.event_id_generator(),
                "type": "response.content_part.added",
                "response_id": self.response_id,
                "item_id": self.item_id,
                "output_index": 0,
                "content_index": 0,
                "part": {
                    "type": "audio",
                    "audio": "",
                    "transcript": ""
                }
            }).model_dump())
        elif mode == "text":
            await self.ws_client.send_json(ResponseContentPartAddedEvent(**{
                "event_id": await self.event_id_generator(),
                "type": "response.content_part.added",
                "response_id": self.response_id,
                "item_id": self.item_id,
                "output_index": 0,
                "content_index": 0,
                "part": {
                   "type": "text",
                    "text": ""
                }
            }).model_dump())


        # mark this as done
        self.conversation_item_created_status = True


    async def update_delta(
        self,
        mode: Literal["text", "transcription", "audio", "function_call"],
        chunk: str | ChoiceDeltaToolCall,
        chunk_in_byte: bytes = None,
    ):
        """ send text chunk to client as well as update internal state """

        # the first delta update will trigger conversation item creation accordingly

        # for function_call mode, this will be sent in the if below where the details of the function call is obtained
        if not self.conversation_item_created_status and mode != "function_call":
            await self.send_conversation_item_created(mode=mode)

        def convert_to_json_string(input_string):
            # Convert the input string to a dictionary
            data_dict = json.loads(input_string)

            # Convert the dictionary to a JSON string with escaped quotation marks
            json_string = json.dumps(data_dict)

            return json_string


        # for overwriting response. Currently use for tool calling
        overwrite_response_dict = {}

        if mode == "text":
            _type = "response.text.delta"
            self.text += chunk
        elif mode == "transcription":
            _type = "response.audio_transcript.delta"
            self.transcription += chunk
        elif mode == "function_call":
            _type = "response.function_call_arguments.delta"
            self.function_calling_arguments = convert_to_json_string(chunk.function.arguments)
            self.function_calling_name = chunk.function.name
            self.function_calling_tool_id = chunk.id
            overwrite_response_dict = {
                "call_id": chunk.id,
                "delta": self.function_calling_arguments,
            }

            # create item
            await self.send_conversation_item_created(
                mode=mode,
                function_name=self.function_calling_name,
                call_id=self.function_calling_tool_id
            )
        elif mode == "audio":
            _type = "response.audio.delta"
            if chunk_in_byte is None:
                raise Exception("chunk_in_byte must be provided for audio mode")

            try:
                # STT generated audio at 24000. However, keep it at 16000 to be consistent with source

                # Convert bytes to numpy array as int16 first (PCM16 format)
                chunk_array = np.frombuffer(chunk_in_byte, dtype=np.int16)

                # Convert to float32 and normalize to [-1, 1]
                chunk_array = chunk_array.astype(np.float32) / 32768.0

                # Ensure chunk_array is 1D
                if chunk_array.ndim == 0:
                    chunk_array = chunk_array.reshape(1)

                # Validate chunk size
                if chunk_array.size == 0:
                    logger.warning("Received empty audio chunk, skipping")
                    return

                # no longer resample here, cause audio in main ws is kept to source
                # Perform sample rate conversion from 24000 Hz to Internal working rate Hz
                # ratio = 16000 / 24000  # Target rate / Source rate
                # chunk_array = samplerate.resample(chunk_array, ratio, 'sinc_best')

                # Concatenate arrays
                if self.audio.size == 0:
                    self.audio = chunk_array
                else:
                    try:
                        # Validate shapes before concatenation
                        if chunk_array.shape[0] == 0:
                            logger.warning("Skipping empty chunk")
                            return

                        self.audio = np.concatenate([self.audio, chunk_array])

                    except Exception as e:
                        logger.error(f"Error concatenating audio chunks: {str(e)}")
                        raise

                # Log some debug information periodically
                if hasattr(self, '_chunk_counter'):
                    self._chunk_counter += 1
                else:
                    self._chunk_counter = 1

                if self._chunk_counter % 10 == 0:
                    logger.debug(f"Audio stats - shape: {self.audio.shape}, "
                                 f"min: {np.min(self.audio)}, max: {np.max(self.audio)}")

            except Exception as e:
                logger.error(f"Error update delta chunk: {str(e)}")
                raise

        # send chunk delta
        response_dict = {
            "event_id": await self.event_id_generator(),
            "type": _type,
            "response_id": self.response_id,
            "item_id": self.item_id,
            "output_index": 0,
            "content_index": 0,
            "delta": chunk  # for audio, send to client base64 string
        }
        if mode == "function_call":
            response_dict.update(overwrite_response_dict)
            response_dict.pop("content_index")

        # logger.info(f"response_dict: {response_dict}")

        await self.ws_client.send_json(ResponseDelta(**response_dict).model_dump(exclude_unset=True))



    async def update_usage(self, usage: UsageResponse):
        """ send usage to client as well as update internal state """
        self.usage = usage

    async def update_content_done(self, content_type: Literal["text", "audio", "transcription", "function_call"]):
        """update client that generation is done"""
        # send done status for one type of content

        # text content
        if content_type=="text":
            await self.ws_client.send_json(ResponseTextDone(**{
                "event_id": await self.event_id_generator(),
                "type": "response.text.done",
                "response_id": self.response_id,
                "item_id": self.item_id,
                "output_index": 0,
                "content_index": 0,
                "text": self.text
            }).model_dump())
            self.text_done = True
        elif content_type=="function_call":
            await self.ws_client.send_json(ResponseFunctionCallArgumentsDone(**{
                "event_id": await self.event_id_generator(),
                "type": "response.function_call_arguments.done",
                "response_id": self.response_id,
                "item_id": self.item_id,
                "output_index": 0,
                "call_id": self.function_calling_tool_id,
                "name": self.function_calling_name,
                "arguments": self.function_calling_arguments,
            }).model_dump())
            self.text_done = True
        elif content_type=="audio":
            await self.ws_client.send_json(ResponseAudioDone(**{
                "event_id": await self.event_id_generator(),
                "type": "response.audio.done",
                "response_id": self.response_id,
                "output_index": 0,
                "content_index": 0,
                "item_id": self.item_id,
            }).model_dump())
            self.audio_done = True

        elif content_type=="transcription":
            await self.ws_client.send_json(ResponseTranscriptDone(**{
                "event_id": await self.event_id_generator(),
                "type": "response.audio_transcript.done",
                "response_id": self.response_id,
                "item_id": self.item_id,
                "output_index": 0,
                "content_index": 0,
                "transcript": self.transcription
            }).model_dump())
            self.transcription_done = True



    async def update_text_or_transcription_task(self, mode: Literal["text", "audio"]):
        content_done_tracker = False
        function_call_tracker = False

        try:
            while not self.cancel_event.is_set():

                if not await self.ws_llm.ensure_connection():
                    raise Exception("Could not establish LLM connection")

                # Receive message from LLM
                message = await asyncio.wait_for(self.ws_llm.connection.recv(), timeout=20)

                # Parse the message
                try:
                    llm_response = json.loads(message)
                    if llm_response.get("object") == "chat.completion.chunk":
                        chat_completion_chunk = ChatCompletionResponse(**llm_response)
                        _delta = chat_completion_chunk.choices[0].delta

                        # finished reason returned with empty delta
                        if _delta.content or (_delta.tool_calls and len(_delta.tool_calls) > 0):
                            text_chunk = _delta.content
                            tool_calls = _delta.tool_calls

                            if text_chunk:      # text answer
                                if mode=="audio":
                                    # put generated text into tts queue for converting to audio
                                    await self.tts_queue.put(text_chunk)

                                await self.update_delta(
                                    mode="transcription" if mode=="audio" else "text",
                                    chunk=text_chunk
                                )
                            elif tool_calls and len(tool_calls) > 0:
                                # for now only return the 1 st function calling even if LLM return multiple

                                if tool_calls[0].function:
                                    await self.update_delta(
                                        mode="function_call",
                                        chunk=tool_calls[0],
                                    )
                                function_call_tracker = True
                            else:
                                logger.error(f"Unknown response type in update_text_or_transcription_task: {llm_response}")

                    elif llm_response.get("type") == "conversation.update.ack":
                        logger.info("Received conversation update acknowledgement from ws_llm")
                    elif llm_response.get("type") == "usage.update":
                        usage_completion = ChatCompletionResponse(**json.loads(llm_response.get("usage"))).usage
                        usage_data = UsageResponseRealTime(
                            input_tokens=usage_completion.prompt_tokens,
                            output_tokens=usage_completion.completion_tokens,
                            total_tokens=usage_completion.total_tokens,
                        )
                        await self.update_usage(usage_data)
                    elif llm_response.get("type") == "generation.complete":
                        await self.tts_queue.put(None)  # mark that text generation completed for TTS
                        if function_call_tracker:
                            _content_type = "function_call"
                        elif mode=="audio":
                            _content_type = "transcription"
                        else:
                            _content_type = "text"

                        await self.update_content_done(content_type=_content_type)
                        await self.content_part_done(generation_type=_content_type)
                        content_done_tracker = True
                        break   # exit the function

                    else:
                        raise Exception(f"Unknown response type in update_text_or_transcription_task: {llm_response}")

                except json.JSONDecodeError:
                    logger.error(f"Failed to parse LLM message to dict: {message}")
                    continue
                except asyncio.TimeoutError:
                    logger.error("LLM connection timed out")
                    continue
                except Exception as e:
                    logger.error(f"Failed to parse LLM message: {e}")
                    raise
        except asyncio.CancelledError:
            logger.info("Text generation task cancelled")
            raise
        except websockets.exceptions.ConnectionClosed:
            logger.error("LLM websocket connection closed unexpectedly")
        except Exception as e:
            logger.error(f"Unexpected error occurred: {str(e)}")


        if not content_done_tracker and not self.cancel_event.is_set():
            logger.error("Text generation task did not complete due to task cancelling")
            await self.tts_queue.put(None)  # mark that text generation completed for TTS

            await self.update_content_done(content_type="transcription" if mode == "audio" else "text")
            await self.content_part_done(generation_type="transcription")
            content_done_tracker = True



    async def send_audio_to_client_task(self):
        """Listen for audio chunks from TTS websocket and forward them to the client"""

        content_done_tracker = False
        try:
            while not self.cancel_event.is_set():
                if not await self.ws_tts.ensure_connection():
                    raise Exception("Could not establish TTS connection")

                try:
                    message = await self.ws_tts.receive_message(timeout=0.05, max_retries=1, disable_warning = True)

                    if isinstance(message, str):
                        try:
                            tts_response = json.loads(message)
                            if tts_response.get("type") == "tts_complete":
                                logger.info("TTS response complete")
                                # mark audio generation as done
                                self.audio_done = True
                                content_done_tracker = True
                                await self.update_content_done(content_type="audio")

                                break   # exit the function
                        except json.JSONDecodeError:
                            logger.warning(f"Received unexpected string message: {message[:100]}...")

                    elif isinstance(message, bytes):
                        try:
                            # Convert to base64 and send
                            audio_base64 = base64.b64encode(message).decode('utf-8')
                            await self.update_delta(
                                mode="audio",
                                chunk=audio_base64,
                                chunk_in_byte=message
                            )
                        except Exception as e:
                            logger.error(f"Error processing audio chunk: {str(e)}")
                            continue


                except asyncio.TimeoutError:
                    # Check if self.function_calling_id is not None
                    if self.function_calling_tool_id is not None:
                        logger.info("Function calling ID is set, considering task completed and no audio needed")
                        # self.audio_done = False
                        # content_done_tracker = False
                        # await self.update_content_done(content_type="audio")
                        break
                    else:
                        continue  # Continue waiting
                except websockets.exceptions.ConnectionClosed:
                    logger.error("TTS websocket connection closed")
                    break
                except asyncio.CancelledError:
                    logger.info("Audio generation task cancelled")
                    raise
                except Exception as e:
                    logger.error(f"Unexpected error occurred in send_audio_to_client_task: {str(e)}")
                    raise

            # check if function calling as well
            if not self.function_calling_tool_id:
                if not content_done_tracker and not self.cancel_event.is_set():
                    logger.info("Audio generation task `did not complete due to task cancelling")
                    self.audio_done = True
                    await self.update_content_done(content_type="audio")
                    content_done_tracker = True

                await self.content_part_done(generation_type="audio")

        except Exception as e:
            logger.error(f"Error in audio_to_client_task: {str(e)}")
            await self.ws_client.send_json({
                "type": "response.failed",
                "response": {
                    "id": self.response_id,
                    "object": "realtime.response",
                    "status": "failed",
                    "last_error": {
                        "code": "internal_error",
                        "message": str(e)
                    }
                }
            })

    async def send_text_to_ws_tts(self):
        """Forward text chunks to TTS service"""
        try:
            # Use the built-in connection management
            if not await self.ws_tts.ensure_connection():
                raise Exception("Could not establish TTS connection")

            while not self.cancel_event.is_set():
                # Get text chunk from queue
                text = await self.tts_queue.get()

                if text is None:  # End of stream marker
                    await self.ws_tts.send_pydantic_message(WSInterTTS(**{
                        "type": "tts.text_done"
                    }))
                    break

                # Send text to TTS service
                success = await self.ws_tts.send_pydantic_message(WSInterTTS(**{
                    "type": "tts.add_text",
                    "text": text
                }))

                if not success:
                    logger.error("Failed to send text to TTS service")
                    break

        except Exception as e:
            logger.error(f"Error in send_text_to_ws_tts: {str(e)}")
        finally:
            # Signal end of text stream
            if self.ws_tts.connection and self.ws_tts.connection.state == State.OPEN:
                await self.ws_tts.send_pydantic_message(WSInterTTS(**{
                    "type": "tts.text_done"
                }))

    async def response_start(self, response_create_request: ResponseCreate, session_config: SessionConfig):
        """ send empty event to client to signal response is created """
        try:
            # send response_create_request to LLM
            request_session_config = response_create_request.response.model_dump() if response_create_request.response else {}

            # find the merged config for this generation
            current_session_config = session_config.merge(request_session_config)
            response_create_request.response = current_session_config


            # at this point, the ws_llm state should already sync with user audio

            await self.ws_llm.send_pydantic_message(response_create_request)


            tasks = []
            if "audio" in current_session_config.modalities:
                text_task = self.update_text_or_transcription_task(mode="audio")
                text_to_ws_tts_task = self.send_text_to_ws_tts()
                audio_to_client_task = self.send_audio_to_client_task()
                tasks = [text_to_ws_tts_task, audio_to_client_task, text_task]
            else:
                text_task = self.update_text_or_transcription_task(mode="text")
                tasks = [text_task]

            await asyncio.gather(*tasks)

        except Exception as e:
            logger.error(f"Error in response_start: {str(e)}")
            raise



import os, sys

from tqdm import tqdm
now_dir = os.getcwd()
sys.path.append(now_dir)

import re
import torch
import LangSegment
from text import chinese
from typing import Dict, List, Tuple
from text.cleaner import clean_text
from text import cleaned_text_to_sequence
from transformers import AutoModelForMaskedLM, AutoTokenizer
from TTS_infer_pack.text_segmentation_method import split_big_text, splits, get_method as get_seg_method

from tools.i18n.i18n import I18nAuto, scan_language_list

language=os.environ.get("language","Auto")
language=sys.argv[-1] if sys.argv[-1] in scan_language_list() else language
i18n = I18nAuto(language=language)
punctuation = set(['!', '?', '…', ',', '.', '-'," "])

def get_first(text:str) -> str:
    pattern = "[" + "".join(re.escape(sep) for sep in splits) + "]"
    text = re.split(pattern, text)[0].strip()
    return text

def merge_short_text_in_array(texts:str, threshold:int) -> list:
    if (len(texts)) < 2:
        return texts
    result = []
    text = ""
    for ele in texts:
        text += ele
        if len(text) >= threshold:
            result.append(text)
            text = ""
    if (len(text) > 0):
        if len(result) == 0:
            result.append(text)
        else:
            result[len(result) - 1] += text
    return result





class TextPreprocessor:
    def __init__(self, bert_model:AutoModelForMaskedLM, 
                 tokenizer:AutoTokenizer, device:torch.device):
        self.bert_model = bert_model
        self.tokenizer = tokenizer
        self.device = device
        
    def preprocess(self, text:str, lang:str, text_split_method:str, version:str="v2")->List[Dict]:
        print(i18n("############ 切分文本 ############"))
        text = self.replace_consecutive_punctuation(text)
        texts = self.pre_seg_text(text, lang, text_split_method)
        result = []
        print(i18n("############ 提取文本Bert特征 ############"))
        for text in tqdm(texts):
            phones, bert_features, norm_text = self.segment_and_extract_feature_for_text(text, lang, version)
            if phones is None or norm_text=="":
                continue
            res={
                "phones": phones,
                "bert_features": bert_features,
                "norm_text": norm_text,
            }
            result.append(res)
        return result

    def pre_seg_text(self, text:str, lang:str, text_split_method:str):
        text = text.strip("\n")
        if len(text) == 0:
            return []
        if (text[0] not in splits and len(get_first(text)) < 4): 
            text = "。" + text if lang != "en" else "." + text
        print(i18n("实际输入的目标文本:"))
        print(text)
        
        seg_method = get_seg_method(text_split_method)
        text = seg_method(text)
        
        while "\n\n" in text:
            text = text.replace("\n\n", "\n")

        _texts = text.split("\n")
        _texts = self.filter_text(_texts)
        _texts = merge_short_text_in_array(_texts, 5)
        texts = []

        
        for text in _texts:
            # 解决输入目标文本的空行导致报错的问题
            if (len(text.strip()) == 0):
               continue
            if not re.sub("\W+", "", text):       
                # 检测一下，如果是纯符号，就跳过。
                continue
            if (text[-1] not in splits): text += "。" if lang != "en" else "."
            
            # 解决句子过长导致Bert报错的问题
            if (len(text) > 510):
                texts.extend(split_big_text(text))
            else:
                texts.append(text)
            
        print(i18n("实际输入的目标文本(切句后):"))
        print(texts)
        return texts
    
    def segment_and_extract_feature_for_text(self, text:str, language:str, version:str="v1")->Tuple[list, torch.Tensor, str]:
        return self.get_phones_and_bert(text, language, version)
        
    def get_phones_and_bert(self, text:str, language:str, version:str, final:bool=False):
        if language in {"en", "all_zh", "all_ja", "all_ko", "all_yue"}:
            language = language.replace("all_","")
            if language == "en":
                LangSegment.setfilters(["en"])
                formattext = " ".join(tmp["text"] for tmp in LangSegment.getTexts(text))
            else:
                # 因无法区别中日韩文汉字,以用户输入为准
                formattext = text
            while "  " in formattext:
                formattext = formattext.replace("  ", " ")
            if language == "zh":
                if re.search(r'[A-Za-z]', formattext):
                    formattext = re.sub(r'[a-z]', lambda x: x.group(0).upper(), formattext)
                    formattext = chinese.mix_text_normalize(formattext)
                    return self.get_phones_and_bert(formattext,"zh",version)
                else:
                    phones, word2ph, norm_text = self.clean_text_inf(formattext, language, version)
                    bert = self.get_bert_feature(norm_text, word2ph).to(self.device)
            elif language == "yue" and re.search(r'[A-Za-z]', formattext):
                    formattext = re.sub(r'[a-z]', lambda x: x.group(0).upper(), formattext)
                    formattext = chinese.mix_text_normalize(formattext)
                    return self.get_phones_and_bert(formattext,"yue",version)
            else:
                phones, word2ph, norm_text = self.clean_text_inf(formattext, language, version)
                bert = torch.zeros(
                    (1024, len(phones)),
                    dtype=torch.float32,
                ).to(self.device)
        elif language in {"zh", "ja", "ko", "yue", "auto", "auto_yue"}:
            textlist=[]
            langlist=[]
            LangSegment.setfilters(["zh","ja","en","ko"])
            if language == "auto":
                for tmp in LangSegment.getTexts(text):
                    langlist.append(tmp["lang"])
                    textlist.append(tmp["text"])
            elif language == "auto_yue":
                for tmp in LangSegment.getTexts(text):
                    if tmp["lang"] == "zh":
                        tmp["lang"] = "yue"
                    langlist.append(tmp["lang"])
                    textlist.append(tmp["text"])
            else:
                for tmp in LangSegment.getTexts(text):
                    if tmp["lang"] == "en":
                        langlist.append(tmp["lang"])
                    else:
                        # 因无法区别中日韩文汉字,以用户输入为准
                        langlist.append(language)
                    textlist.append(tmp["text"])
            # print(textlist)
            # print(langlist)
            phones_list = []
            bert_list = []
            norm_text_list = []
            for i in range(len(textlist)):
                lang = langlist[i]
                phones, word2ph, norm_text = self.clean_text_inf(textlist[i], lang, version)
                bert = self.get_bert_inf(phones, word2ph, norm_text, lang)
                phones_list.append(phones)
                norm_text_list.append(norm_text)
                bert_list.append(bert)
            bert = torch.cat(bert_list, dim=1)
            phones = sum(phones_list, [])
            norm_text = ''.join(norm_text_list)

        if not final and len(phones) < 6:
            return self.get_phones_and_bert("." + text,language,version,final=True)

        return phones, bert, norm_text


    def get_bert_feature(self, text:str, word2ph:list)->torch.Tensor:
        with torch.no_grad():
            inputs = self.tokenizer(text, return_tensors="pt")
            for i in inputs:
                inputs[i] = inputs[i].to(self.device)
            res = self.bert_model(**inputs, output_hidden_states=True)
            res = torch.cat(res["hidden_states"][-3:-2], -1)[0].cpu()[1:-1]
        assert len(word2ph) == len(text)
        phone_level_feature = []
        for i in range(len(word2ph)):
            repeat_feature = res[i].repeat(word2ph[i], 1)
            phone_level_feature.append(repeat_feature)
        phone_level_feature = torch.cat(phone_level_feature, dim=0)
        return phone_level_feature.T
    
    def clean_text_inf(self, text:str, language:str, version:str="v2"):
        phones, word2ph, norm_text = clean_text(text, language, version)
        phones = cleaned_text_to_sequence(phones, version)
        return phones, word2ph, norm_text

    def get_bert_inf(self, phones:list, word2ph:list, norm_text:str, language:str):
        language=language.replace("all_","")
        if language == "zh":
            feature = self.get_bert_feature(norm_text, word2ph).to(self.device)
        else:
            feature = torch.zeros(
                (1024, len(phones)),
                dtype=torch.float32,
            ).to(self.device)

        return feature


    def filter_text(self,texts):
        _text=[]
        if all(text in [None, " ", "\n",""] for text in texts):
            raise ValueError(i18n("请输入有效文本"))
        for text in texts:
            if text in  [None, " ", ""]:
                pass
            else:
                _text.append(text)
        return _text
    

    def replace_consecutive_punctuation(self,text):
        punctuations = ''.join(re.escape(p) for p in punctuation)
        pattern = f'([{punctuations}])([{punctuations}])+'
        result = re.sub(pattern, r'\1', text)
        return result






# modified from https://github.com/yangdongchao/SoundStorm/blob/master/soundstorm/s1/AR/data/dataset.py
# reference: https://github.com/lifeiteng/vall-e
import pdb
import sys

# sys.path.append("/data/docker/liujing04/gpt-vits/mq-vits-s1bert_no_bert")
import traceback, os
from typing import Dict
from typing import List

import numpy as np
import pandas as pd
import torch, json
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from transformers import AutoTokenizer

version = os.environ.get('version',None)

from text import cleaned_text_to_sequence

# from config import exp_dir


def batch_sequences(sequences: List[np.array], axis: int = 0, pad_value: int = 0):
    seq = sequences[0]
    ndim = seq.ndim
    if axis < 0:
        axis += ndim
    dtype = seq.dtype
    pad_value = dtype.type(pad_value)
    seq_lengths = [seq.shape[axis] for seq in sequences]
    max_length = np.max(seq_lengths)

    padded_sequences = []
    for seq, length in zip(sequences, seq_lengths):
        padding = (
            [(0, 0)] * axis + [(0, max_length - length)] + [(0, 0)] * (ndim - axis - 1)
        )
        padded_seq = np.pad(seq, padding, mode="constant", constant_values=pad_value)
        padded_sequences.append(padded_seq)
    batch = np.stack(padded_sequences)
    return batch


class Text2SemanticDataset(Dataset):
    """dataset class for text tokens to semantic model training."""

    def __init__(
        self,
        phoneme_path: str,
        semantic_path: str,
        max_sample: int = None,
        max_sec: int = 100,
        pad_val: int = 1024,
        # min value of phoneme/sec
        min_ps_ratio: int = 3,
        # max value of phoneme/sec
        max_ps_ratio: int = 25,
    ) -> None:
        super().__init__()

        self.semantic_data = pd.read_csv(
            semantic_path, delimiter="\t", encoding="utf-8"
        )
        # get dict
        self.path2 = phoneme_path  # "%s/2-name2text.txt"%exp_dir#phoneme_path
        self.path3 = "%s/3-bert" % (
            os.path.dirname(phoneme_path)
        )  # "%s/3-bert"%exp_dir#bert_dir
        self.path6 = semantic_path  # "%s/6-name2semantic.tsv"%exp_dir#semantic_path
        assert os.path.exists(self.path2)
        assert os.path.exists(self.path6)
        self.phoneme_data = {}
        with open(self.path2, "r", encoding="utf8") as f:
            lines = f.read().strip("\n").split("\n")

        for line in lines:
            tmp = line.split("\t")
            if len(tmp) != 4:
                continue
            self.phoneme_data[tmp[0]] = [tmp[1], tmp[2], tmp[3]]

        # self.phoneme_data = np.load(phoneme_path, allow_pickle=True).item()
        # pad for semantic tokens
        self.PAD: int = pad_val
        # self.hz = 25
        # with open("/data/docker/liujing04/gpt-vits/mq-vits-s1bert_no_bert/configs/s2.json", "r") as f:data = f.read()
        # data=json.loads(data)["model"]["semantic_frame_rate"]#50hz
        # self.hz=int(data[:-2])#
        self.hz = int(os.environ.get("hz", "25hz")[:-2])

        # max seconds of semantic token
        self.max_sec = max_sec
        self.min_ps_ratio = min_ps_ratio
        self.max_ps_ratio = max_ps_ratio

        if max_sample is not None:
            self.semantic_data = self.semantic_data[:max_sample]

        # {idx: (semantic, phoneme)}
        # semantic list, phoneme list
        self.semantic_phoneme = []
        self.item_names = []

        self.inited = False

        if not self.inited:
            # 调用初始化函数
            self.init_batch()
            self.inited = True
            del self.semantic_data
            del self.phoneme_data
        # self.tokenizer = AutoTokenizer.from_pretrained("hfl/chinese-roberta-wwm-ext-large")
        # self.tokenizer = AutoTokenizer.from_pretrained("/data/docker/liujing04/bert-vits2/Bert-VITS2-master20231106/bert/chinese-roberta-wwm-ext-large")

    def init_batch(self):
        semantic_data_len = len(self.semantic_data)
        phoneme_data_len = len(self.phoneme_data.keys())
        print("semantic_data_len:", semantic_data_len)
        print("phoneme_data_len:", phoneme_data_len)
        print(self.semantic_data)
        idx = 0
        num_not_in = 0
        num_deleted_bigger = 0
        num_deleted_ps = 0
        for i in range(semantic_data_len):
            # 先依次遍历
            # get str
            item_name = self.semantic_data.iloc[i,0]
            # print(self.phoneme_data)
            try:
                phoneme, word2ph, text = self.phoneme_data[item_name]
            except Exception:
                traceback.print_exc()
                # print(f"{item_name} not in self.phoneme_data !")
                num_not_in += 1
                continue

            semantic_str = self.semantic_data.iloc[i,1]
            # get token list
            semantic_ids = [int(idx) for idx in semantic_str.split(" ")]
            # (T), 是否需要变成 (1, T) -> 不需要，因为需要求 len
            # 过滤掉太长的样本
            if (
                len(semantic_ids) > self.max_sec * self.hz
            ):  #########1###根据token个数推测总时长过滤时长60s（config里）#40*25=1k
                num_deleted_bigger += 1
                continue
            # (T, ), 这个速度不会很慢，所以可以在一开始就处理，无需在 __getitem__ 里面单个处理####
            phoneme = phoneme.split(" ")

            try:
                phoneme_ids = cleaned_text_to_sequence(phoneme, version)
            except:
                traceback.print_exc()
                # print(f"{item_name} not in self.phoneme_data !")
                num_not_in += 1
                continue
            # if len(phoneme_ids) >400:###########2：改为恒定限制为semantic/2.5就行
            if (
                len(phoneme_ids) > self.max_sec * self.hz / 2.5
            ):  ###########2：改为恒定限制为semantic/2.5就行
                num_deleted_ps += 1
                continue
            # if len(semantic_ids) > 1000:###########3
            #     num_deleted_bigger += 1
            #     continue

            ps_ratio = len(phoneme_ids) / (len(semantic_ids) / self.hz)

            if (
                ps_ratio > self.max_ps_ratio or ps_ratio < self.min_ps_ratio
            ):  ##########4#3~25#每秒多少个phone
                num_deleted_ps += 1
                # print(item_name)
                continue

            self.semantic_phoneme.append((semantic_ids, phoneme_ids))
            idx += 1
            self.item_names.append(item_name)

        min_num = 100  # 20直接不补#30补了也不存ckpt
        leng = len(self.semantic_phoneme)
        if leng < min_num:
            tmp1 = self.semantic_phoneme
            tmp2 = self.item_names
            self.semantic_phoneme = []
            self.item_names = []
            for _ in range(max(2, int(min_num / leng))):
                self.semantic_phoneme += tmp1
                self.item_names += tmp2
        if num_not_in > 0:
            print(f"there are {num_not_in} semantic datas not in phoneme datas")
        if num_deleted_bigger > 0:
            print(
                f"deleted {num_deleted_bigger} audios who's duration are bigger than {self.max_sec} seconds"
            )
        if num_deleted_ps > 0:
            # 4702 for LibriTTS, LirbriTTS 是标注数据, 是否需要筛？=> 需要，有值为 100 的极端值
            print(
                f"deleted {num_deleted_ps} audios who's phoneme/sec are bigger than {self.max_ps_ratio} or smaller than {self.min_ps_ratio}"
            )
        """
        there are 31 semantic datas not in phoneme datas
        deleted 34 audios who's duration are bigger than 54 seconds
        deleted 3190 audios who's phoneme/sec are bigger than 25 or smaller than 3
        dataset.__len__(): 366463

        """
        # 345410 for LibriTTS
        print("dataset.__len__():", self.__len__())

    def __get_item_names__(self) -> List[str]:
        return self.item_names

    def __len__(self) -> int:
        return len(self.semantic_phoneme)

    def __getitem__(self, idx: int) -> Dict:
        semantic_ids, phoneme_ids = self.semantic_phoneme[idx]
        item_name = self.item_names[idx]
        phoneme_ids_len = len(phoneme_ids)
        # semantic tokens target
        semantic_ids_len = len(semantic_ids)

        flag = 0
        path_bert = "%s/%s.pt" % (self.path3, item_name)
        if os.path.exists(path_bert) == True:
            bert_feature = torch.load(path_bert, map_location="cpu")
        else:
            flag = 1
        if flag == 1:
            # bert_feature=torch.zeros_like(phoneme_ids,dtype=torch.float32)
            bert_feature = None
        else:
            assert bert_feature.shape[-1] == len(phoneme_ids)
        return {
            "idx": idx,
            "phoneme_ids": phoneme_ids,
            "phoneme_ids_len": phoneme_ids_len,
            "semantic_ids": semantic_ids,
            "semantic_ids_len": semantic_ids_len,
            "bert_feature": bert_feature,
        }

    def get_sample_length(self, idx: int):
        semantic_ids = self.semantic_phoneme[idx][0]
        sec = 1.0 * len(semantic_ids) / self.hz
        return sec

    def collate(self, examples: List[Dict]) -> Dict:
        sample_index: List[int] = []
        phoneme_ids: List[torch.Tensor] = []
        phoneme_ids_lens: List[int] = []
        semantic_ids: List[torch.Tensor] = []
        semantic_ids_lens: List[int] = []
        # return

        for item in examples:
            sample_index.append(item["idx"])
            phoneme_ids.append(np.array(item["phoneme_ids"], dtype=np.int64))
            semantic_ids.append(np.array(item["semantic_ids"], dtype=np.int64))
            phoneme_ids_lens.append(item["phoneme_ids_len"])
            semantic_ids_lens.append(item["semantic_ids_len"])

        # pad 0
        phoneme_ids = batch_sequences(phoneme_ids)
        semantic_ids = batch_sequences(semantic_ids, pad_value=self.PAD)

        # # convert each batch to torch.tensor
        phoneme_ids = torch.tensor(phoneme_ids)
        semantic_ids = torch.tensor(semantic_ids)
        phoneme_ids_lens = torch.tensor(phoneme_ids_lens)
        semantic_ids_lens = torch.tensor(semantic_ids_lens)
        bert_padded = torch.FloatTensor(len(examples), 1024, max(phoneme_ids_lens))
        bert_padded.zero_()

        for idx, item in enumerate(examples):
            bert = item["bert_feature"]
            if bert != None:
                bert_padded[idx, :, : bert.shape[-1]] = bert

        return {
            # List[int]
            "ids": sample_index,
            # torch.Tensor (B, max_phoneme_length)
            "phoneme_ids": phoneme_ids,
            # torch.Tensor (B)
            "phoneme_ids_len": phoneme_ids_lens,
            # torch.Tensor (B, max_semantic_ids_length)
            "semantic_ids": semantic_ids,
            # torch.Tensor (B)
            "semantic_ids_len": semantic_ids_lens,
            # torch.Tensor (B, 1024, max_phoneme_length)
            "bert_feature": bert_padded,
        }


if __name__ == "__main__":
    root_dir = "/data/docker/liujing04/gpt-vits/prepare/dump_mix/"
    dataset = Text2SemanticDataset(
        phoneme_path=root_dir + "phoneme_train.npy",
        semantic_path=root_dir + "semantic_train.tsv",
    )

    batch_size = 12
    dataloader = DataLoader(
        dataset, batch_size=batch_size, collate_fn=dataset.collate, shuffle=False
    )
    for i, batch in enumerate(dataloader):
        if i % 1000 == 0:
            print(i)
        # if i == 0:
        #     print('batch["ids"]:', batch["ids"])
        # print('batch["phoneme_ids"]:', batch["phoneme_ids"],
        #       batch["phoneme_ids"].shape)
        # print('batch["phoneme_ids_len"]:', batch["phoneme_ids_len"],
        #       batch["phoneme_ids_len"].shape)
        # print('batch["semantic_ids"]:', batch["semantic_ids"],
        #       batch["semantic_ids"].shape)
        # print('batch["semantic_ids_len"]:', batch["semantic_ids_len"],
        #       batch["semantic_ids_len"].shape)



from . import cli
from gallama.logger.logger import logger
from gallama.backend.llm.thinking_template import THINKING_TEMPLATE, Thinking
from . import app
from .app import make_server
from .config import ConfigManager
from .dependencies import model_manager
from .dependencies_server import get_server_manager, get_server_logger, DEFAULT_ZMQ_URL


# using absolute import here as this file will be run alone
import sys
from fastapi import FastAPI, Request, APIRouter
from fastapi.middleware.cors import CORSMiddleware
from gallama.data_classes import (
    ModelSpec,
    GenQueue,
    GenQueueDynamic
)
import signal
import argparse
import uvicorn
from fastapi.exceptions import RequestValidationError
import json
from fastapi.responses import JSONResponse
from gallama.utils import parse_request_body
from gallama.config.config_manager import ConfigManager
from gallama.data import ARTIFACT_SYSTEM_PROMPT
import os
from contextlib import asynccontextmanager
from logging import DEBUG
from gallama.logger.logger import get_logger
import base64
from gallama.dependencies import get_model_manager
from gallama.routes import (
    chat_router,
    embedding_router,
    model_management_router,
    audio_router,
    ws_stt_router,
    ws_llm_router,
    ws_tts_router,
    ws_video_router
)
from tempfile import SpooledTemporaryFile

# Add this after your imports to clear logging from 3rd party module

#streaming example
#https://blog.gopenai.com/how-to-stream-output-in-llm-based-applications-to-boost-user-experience-e9fcf582777a

logger = get_logger()

# add endpoint
router = APIRouter()
router.include_router(chat_router)
router.include_router(embedding_router)
router.include_router(model_management_router)
router.include_router(audio_router)
router.include_router(ws_stt_router)
router.include_router(ws_llm_router)
router.include_router(ws_tts_router)
router.include_router(ws_video_router)


config_manager = ConfigManager()
model_manager = get_model_manager()
model_ready = False

@router.get("/")
async def read_root():
    return {"Hello": "PAI"}


@router.options("/v1/chat/completions")
async def options_handler(request: Request):
    return JSONResponse(
        status_code=200,
        content={"message": "Local OpenAI response for OPTIONS request"},
        headers={
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Methods": "GET, POST, OPTIONS",
            "Access-Control-Allow-Headers": "",
        },
    )



async def startup_event():
    # run some dummy generation so that cache is initialized
    logger.info("Generator initialization")

    # warm up LLM
    if model_manager.llm_dict:
        gen_queue = GenQueue()
        for _model_name, _model in model_manager.llm_dict.items():
            await _model.chat_raw(
                prompt=f"{ARTIFACT_SYSTEM_PROMPT}\nWrite a 500 words story on Llama",
                # stream=False,
                max_tokens=50,
                gen_queue=gen_queue,
                quiet=True,
                request=None,
            )

            logger.info(f"LLM| {_model_name} | warmed up")
        gen_queue = None

    if model_manager.stt_dict:
        logger.info("STT warmed up NOT YET IMPLEMENTED")

    if model_manager.tts_dict:
        for _model_name, tts in model_manager.tts_dict.items():
            _, _ = await tts.text_to_speech(
                text="hello",
                stream=False,
                batching=False,
                batch_size=1
            )
            logger.info(f"TTS| {_model_name} | warmed up")

    if model_manager.embedding_dict:
        logger.info("Embedding warmed up NOT YET IMPLEMENTED")



@asynccontextmanager
async def lifespan(app: FastAPI):
    # Startup code
    await startup_event()
    try:
        yield
    finally:
        # Cleanup code
        logger.info("Cleaning up ZMQ connections...")
        for handler in logger.handlers:
            if hasattr(handler, 'close'):
                handler.close()


def make_server(args):
    global logger
    global draft_spec_dict

    # Add signal handlers for graceful shutdown
    def signal_handler(signum, frame):
        logger.info("Received shutdown signal, cleaning up...")
        for handler in logger.handlers:
            if hasattr(handler, 'close'):
                handler.close()
        sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    # load yaml file of model info
    logger.info(args)
    model_dict = {}
    draft_spec_dict = {}    # for speculative decoding

    # parse cli argument to get model info
    # Handle model specification
    model_spec = None
    if args.model_spec:
        try:
            # Decode and parse the ModelSpec from base64 JSON
            model_json = base64.b64decode(args.model_spec).decode('utf-8')
            model_spec = ModelSpec.model_validate_json(model_json)
        except Exception as e:
            logger.error(f"Failed to parse model specification: {e}")
            sys.exit(1)

        # set the env for log to zmq
        # Note: This assumes you want to use the first model's ID if multiple are provided
        os.environ['MODEL_NAME'] = model_spec.model_name
        os.environ['MODEL_PORT'] = str(args.port)

    # set logger level
    if args.verbose:
        logger.setLevel(DEBUG)
        os.environ["LOCAL_OPEN_AI_VERBOSE"] = '2'   # turn on verbosity for all

    args.model_spec = model_spec
    logger.info("Parsed Arguments:" + str(args))  # Debug statement

    if args.detached:
        # send logging to zmq so that it will show in the parent log
        logger = get_logger(name="child", to_zmq=True, to_console=False)
    else:
        # keep the default logger declared on top
        pass

    app = FastAPI(lifespan=lifespan)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"], #origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    app.include_router(router)

    @app.middleware("http")
    @app.middleware("https")
    async def log_requests(request: Request, call_next):
        try:
            # if request.method in ("POST", "PUT", "PATCH"):
            #     # Parse body and preserve it
            #     body_content, is_multipart = await parse_request_body(request)
            #
            #     # Handle file uploads explicitly
            #     if is_multipart or isinstance(body_content, SpooledTemporaryFile):
            #         request_content = "File upload content not logged"
            #     elif isinstance(body_content, dict):
            #         request_content = json.dumps(body_content, indent=2)
            #     elif isinstance(body_content, str):
            #         try:
            #             parsed_content = json.loads(body_content)
            #             request_content = json.dumps(parsed_content, indent=2)
            #         except json.JSONDecodeError:
            #             request_content = body_content
            #     else:
            #         request_content = str(body_content)
            #
            #     logger.debug(
            #         f"API Request:\n"
            #         f"Method: {request.method}\n"
            #         f"URL: {request.url}\n"
            #         f"Headers: {dict(request.headers)}\n"
            #         f"Content-Type: {request.headers.get('content-type', 'Not specified')}\n"
            #         f"Body: {request_content if request_content else 'No Body'}"
            #     )

            # Proceed with the request
            response = await call_next(request)
            return response

        except Exception as e:
            logger.error(f"Middleware error: {str(e)}", exc_info=True)
            return JSONResponse(
                status_code=500,
                content={"detail": "Internal server error in middleware"}
            )

    if model_spec:
        # load model
        model_manager.load_model(model_spec)
        model_manager.model_ready = True

    uvicorn.run(app, host=args.host, port=args.port, log_config=None)    # disable log_config to use our custom logger


def parse_dict(arg):
    """Parses a key=value string and returns a dictionary."""
    result = {}
    for pair in arg.split():
        key, value = pair.split('=')
        result[key] = value.strip("'")  # Strip single quotes here as well
    return result


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser(description="Launch local AI model")
    arg_parser.add_argument("--model-spec", type=str, help="Base64 encoded JSON ModelSpec object")
    arg_parser.add_argument('-v', "--verbose", action='store_true', help="Turn on more verbose logging")
    arg_parser.add_argument('-d', "--detached", action='store_true', help="Log to ZeroMQ")
    arg_parser.add_argument("--host", type=str, default="127.0.0.1", help="The host to bind to.")
    arg_parser.add_argument('-p', "--port", type=int, default=8000, help="The port to bind to.")
    arg_parser.add_argument("--reload", action="store_true", help="Enable auto-reload.")

    args = arg_parser.parse_args()

    make_server(args)



import logging
from logging.config import dictConfig
from logging.handlers import RotatingFileHandler
import os
import sys
import textwrap
from colorama import Fore, Back, Style, init
import json
import re
from pygments import highlight
from pygments.lexers import PythonLexer, MarkdownLexer, JsonLexer
from pygments.formatters import TerminalFormatter
from pygments.token import Token
from pygments.lexer import RegexLexer
from pydantic import BaseModel, validator
from typing import Optional, Union
import zmq

# Initialize colorama
init(autoreset=True)

# Aggressively reset all logging
logging.root.handlers = []
logging.root.setLevel(logging.NOTSET)


DEFAULT_ZMQ_URL = "tcp://127.0.0.1:5555"  # Using 5559 as a standard port for logging


class PromptLexer(RegexLexer):
    SPECIAL_TOKENS = [
        '[INST]', '[/INST]',
        '<s>', '</s>',
        '[AVAILABLE_TOOLS]', '[/AVAILABLE_TOOLS]', '[TOOL_RESULTS]', '[/TOOL_RESULTS]',
        '<|begin_of_text|>', '<|end_of_text|>', '<|start_header_id|>', '<|end_header_id|>', '<|eot_id|>',
        '<|assistant|>', '<|user|>', '<|end|>',
        '<|im_start|>', '<|im_end|>',
        '<start_of_turn>', '<end_of_turn>',
    ]

    tokens = {
        'root': [
            (r'|'.join(map(re.escape, SPECIAL_TOKENS)), Token.Keyword),
            (r'.*?\n', Token.Text),
        ]
    }


class ColorTabularFormatter(logging.Formatter):
    def __init__(self, max_width=50):
        super().__init__()
        self.max_width = max_width
        self.COLORS = {
            'DEBUG': Fore.CYAN,
            'INFO': Fore.GREEN,
            'WARNING': Fore.YELLOW,
            'ERROR': Fore.RED,
            'CRITICAL': Fore.RED + Back.WHITE
        }
        self.python_lexer = PythonLexer()
        self.markdown_lexer = MarkdownLexer()
        self.json_lexer = JsonLexer()
        self.prompt_lexer = PromptLexer()
        self.terminal_formatter = TerminalFormatter()

    def format(self, record):
        level = record.levelname
        filename = record.filename
        message = record.getMessage()

        color = self.COLORS.get(level, '')
        reset = Style.RESET_ALL

        message = self.highlight_content(message)

        message_lines = message.splitlines()

        formatted_lines = []
        for i, line in enumerate(message_lines):
            wrapped_lines = textwrap.wrap(line, self.max_width)

            for j, wrapped_line in enumerate(wrapped_lines):
                if i == 0 and j == 0:
                    formatted_lines.append(
                        f"{color}{level:<8}{reset} | {wrapped_line:<{self.max_width}} | {Fore.BLUE}{filename}{reset}")
                else:
                    formatted_lines.append(f"{' ':8}   {wrapped_line:<{self.max_width}}   ")

        record.plain_message = record.getMessage()

        return "\n".join(formatted_lines)

    def highlight_content(self, message):
        def highlight_code(match):
            lang = match.group(1) or 'python'
            code = match.group(2)

            if lang in ('python', 'py', ''):
                highlighted = highlight(code, self.python_lexer, self.terminal_formatter)
            elif lang in ('md', 'markdown'):
                highlighted = highlight(code, self.markdown_lexer, self.terminal_formatter)
            elif lang == 'json':
                try:
                    parsed = json.loads(code)
                    formatted = json.dumps(parsed, indent=2)
                    highlighted = highlight(formatted, self.json_lexer, self.terminal_formatter)
                except json.JSONDecodeError:
                    highlighted = code
            else:
                highlighted = highlight(code, self.python_lexer, self.terminal_formatter)

            return f"```{lang}\n{highlighted}```"

        pattern = r'```(\w*)\n(.*?)```'
        message = re.sub(pattern, highlight_code, message, flags=re.DOTALL)

        highlighted_prompt = highlight(message, self.prompt_lexer, self.terminal_formatter)
        return highlighted_prompt


class PlainTextFormatter(logging.Formatter):
    def format(self, record):
        return record.plain_message if hasattr(record, 'plain_message') else record.getMessage()


class ZeroMQHandler(logging.Handler):
    def __init__(self, zmq_url=DEFAULT_ZMQ_URL):
        super().__init__()
        self.zmq_url = zmq_url
        self._context = None
        self._socket = None
        self._initialize_socket()

    def _initialize_socket(self):
        """Initialize or reinitialize the ZMQ socket"""
        try:
            if self._socket:
                self._socket.close()
            if self._context:
                self._context.term()

            self._context = zmq.Context()
            self._socket = self._context.socket(zmq.PUSH)
            self._socket.setsockopt(zmq.LINGER, 0)
            self._socket.connect(self.zmq_url)
        except Exception as e:
            print(f"Error initializing ZMQ socket: {e}")
            self._socket = None
            self._context = None

    def emit(self, record):
        if not self._socket:
            self._initialize_socket()
            if not self._socket:
                return  # Skip logging if socket initialization failed

        log_entry = self.format(record)
        try:
            self._socket.send_json({
                'log': log_entry,
                'level': record.levelname,
                'model': os.environ.get('MODEL_NAME', 'unknown'),
                'port': os.environ.get('MODEL_PORT', 'unknown')
            }, flags=zmq.NOBLOCK)
        except zmq.Again:
            print("Warning: ZMQ socket buffer full, log message dropped")
        except zmq.error.ZMQError as e:
            print(f"ZMQ Error in emit: {e}")
            self._initialize_socket()  # Try to reinitialize socket
        except Exception as e:
            print(f"Error in ZeroMQHandler: {e}")
            self.handleError(record)

    def close(self):
        """Properly close the ZMQ socket and context"""
        try:
            if self._socket:
                self._socket.close()
            if self._context:
                self._context.term()
        except Exception as e:
            print(f"Error closing ZMQ handler: {e}")
        finally:
            self._socket = None
            self._context = None
            super().close()


class LogConfig(BaseModel):
    LOGGER_NAME: str = "logger"
    LOG_LEVEL: Union[str, int] = "DEBUG"
    LOG_FILE: Optional[str] = None
    ZMQ_URL: Optional[str] = None
    TO_CONSOLE: bool = True
    TO_FILE: bool = False
    TO_ZMQ: bool = False
    MAX_FILE_SIZE: int = 10 * 1024 * 1024  # 10 MB
    BACKUP_COUNT: int = 5

    version: int = 1
    disable_existing_loggers: bool = False
    formatters: dict = {
        "color_tabular": {
            "()": ColorTabularFormatter,
            "max_width": 250,
        },
        "plain": {
            "()": PlainTextFormatter,
        },
    }
    handlers: dict = {}
    loggers: dict = {}

    @validator('LOG_FILE')
    def validate_log_file(cls, v):
        if v is not None:
            if not ensure_dir_exists(v):
                raise ValueError(f"Cannot create directory for log file: {v}")
        return v



def ensure_dir_exists(file_path):
    directory = os.path.dirname(file_path)
    if directory and not os.path.exists(directory):
        try:
            os.makedirs(directory, exist_ok=True)
        except PermissionError:
            print(f"Error: Permission denied when trying to create directory {directory}")
            return False
    return True


def setup_logger(config: LogConfig):
    handlers = []

    if config.TO_CONSOLE:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(ColorTabularFormatter(max_width=250))
        handlers.append(console_handler)

    if config.TO_FILE and config.LOG_FILE:
        if ensure_dir_exists(config.LOG_FILE):
            file_handler = RotatingFileHandler(
                filename=config.LOG_FILE,
                maxBytes=config.MAX_FILE_SIZE,
                backupCount=config.BACKUP_COUNT
            )
            file_handler.setFormatter(PlainTextFormatter())
            handlers.append(file_handler)
        else:
            print(f"Warning: Unable to create log file {config.LOG_FILE}. File logging will be disabled.")
            config.TO_FILE = False

    if config.TO_ZMQ and config.ZMQ_URL:
        zmq_handler = ZeroMQHandler(zmq_url=config.ZMQ_URL)
        zmq_handler.setFormatter(PlainTextFormatter())
        handlers.append(zmq_handler)

    logger = logging.getLogger(config.LOGGER_NAME)
    logger.setLevel(config.LOG_LEVEL)
    logger.handlers = []  # Remove any existing handlers
    for handler in handlers:
        logger.addHandler(handler)

    return logger

def get_logger(
    name='logger',
    log_file="./log/llm_response.log",
    log_level=None,
    to_console=True,
    to_file=False,
    to_zmq=True,
    zmq_url=DEFAULT_ZMQ_URL
):
    # Set verbosity
    logger_verbose = os.getenv('LOCAL_OPEN_AI_VERBOSE', '1')

    if not log_level:
        if logger_verbose == '1':
            log_level = logging.INFO
        elif logger_verbose == '2':
            log_level = logging.DEBUG

    config = LogConfig(
        LOGGER_NAME=name,
        LOG_LEVEL=log_level,
        LOG_FILE=log_file,
        ZMQ_URL=zmq_url,
        TO_CONSOLE=to_console,
        TO_FILE=to_file,
        TO_ZMQ=to_zmq
    )
    return setup_logger(config)

logger = get_logger()


from .hypothesis import HypothesisBuffer
from .base import ASRBase
import numpy as np
from typing import Union, BinaryIO, Optional
import asyncio
from concurrent.futures import ThreadPoolExecutor
import os
from datetime import datetime
import soundfile as sf
from ...data_classes import TranscriptionResponse, LanguageType
from ...data_classes.realtime_client_proto import TurnDetectionConfig
from .audio_buffer import AudioBufferWithTiming
from .vad import VADProcessor
from gallama.logger import logger


class ASRProcessor:
    """
    ASRProcessor processes audio buffers for real-time ASR (Automatic Speech Recognition).
    It manages an audio buffer, transcribes audio using an ASR backend, and trims processed portions
    to maintain efficiency. Supports trimming based on sentences or audio segments.
    """

    SAMPLING_RATE = 16000  # Standard audio sampling rate

    def __init__(
        self,
        asr: ASRBase,
        tokenizer=None,
        buffer_trimming=("segment", 15),
        min_context_needed=3.0,  # require 3 second for a content to be process -> higher number better accuracy
        debug_audio_dir=None,  # New parameter for debug audio directory
        vad_config: Optional[TurnDetectionConfig] = None,
    ):
        """
        Initializes the ASR processor.

        Args:
            asr: ASR backend object for transcribing audio.
            tokenizer: Sentence tokenizer for the target language. Required if trimming mode is "sentence".
            buffer_trimming: Tuple (mode, duration).
                             - mode: "sentence" or "segment" (trimming method).
                             - duration: Max buffer length (in seconds) before trimming occurs.
            debug_audio_dir: Directory to store debug audio files
        """
        self.asr = asr
        self.tokenizer = tokenizer
        self.debug_audio_dir = debug_audio_dir
        self.min_context_needed = min_context_needed

        if debug_audio_dir and not os.path.exists(debug_audio_dir):
            os.makedirs(debug_audio_dir)

        self.trimming_mode, self.trimming_duration = buffer_trimming

        # Initialize two separate buffers
        self.audio_buffer = AudioBufferWithTiming(sample_rate=self.SAMPLING_RATE)
        self.vad_audio_buffer = AudioBufferWithTiming(sample_rate=self.SAMPLING_RATE)

        self.transcript_buffer = HypothesisBuffer()
        self.transcript_buffer.last_committed_time = 0
        self.committed_transcriptions = []
        self.buffer_count = 0

        self.vad_config = vad_config if vad_config else TurnDetectionConfig()
        self.vad_enable: bool = self.vad_config.create_response

        self.vad = None
        if self.vad_enable:
            self.initialize_vad(self.vad_config)

        self.vad_speech_active = False
        self.vad_start_sent = False
        self.vad_end_sent = False

    def add_audio_chunk(self, audio_chunk):
        """
        Adds an audio chunk to the appropriate buffer based on VAD mode.
        """
        if self.vad_enable:
            self.vad_audio_buffer.add_chunk(audio_chunk)
        else:
            self.audio_buffer.add_chunk(audio_chunk)

    def reset_buffers_for_mode_change(self, new_vad_enable: bool):
        """
        Reset the appropriate buffer when switching between VAD and non-VAD modes.
        """
        if new_vad_enable and not self.vad_enable:
            # Switching from non-VAD to VAD mode: reset the VAD buffer
            logger.info("Switching to VAD mode: resetting VAD buffer")
            self.vad_audio_buffer.reset()
        elif not new_vad_enable and self.vad_enable:
            # Switching from VAD to non-VAD mode: reset the non-VAD buffer
            logger.info("Switching to non-VAD mode: resetting non-VAD buffer")
            self.audio_buffer.reset()

        # Update the VAD enable flag
        self.vad_enable = new_vad_enable


    def process_audio(self, is_final: bool = False):
        """
        Processes the current audio buffer, transcribes it, and handles buffer trimming.
        """
        if self.vad_enable:
            buffer = self.vad_audio_buffer
        else:
            buffer = self.audio_buffer

        if len(buffer) == 0:
            logger.debug("Audio buffer is empty, returning early")
            return None, None, "", []

        buffer.mark_processing_start()

        vad_events = []
        speech_start_ms = None
        speech_end_ms = None

        if self.vad_enable and self.vad:
            current_offset = buffer.last_processed_sample_vad
            vad_result_start, vad_result_end = self.vad.process_chunk(buffer, current_offset)

            logger.info(f"VAD result_start: {vad_result_start}")
            logger.info(f"VAD result_end: {vad_result_end}")

            if vad_result_start and vad_result_start['speech_detected'] and not self.vad_speech_active and not self.vad_start_sent:
                self.vad_speech_active = True
                self.vad_start_sent = True
                self.vad_end_sent = False
                speech_start_ms = vad_result_start.get('start_time', 0.0)
                vad_events.append({
                    'type': 'start',
                    'timestamp_ms': int(speech_start_ms),
                    'confidence': vad_result_start.get('confidence', 0.0)
                })
                logger.info(f"Speech start event queued at {speech_start_ms}ms")

            if vad_result_end and (vad_result_end['speech_ended'] or (is_final and self.vad_speech_active)) and not self.vad_end_sent:
                is_final = True
                self.vad_speech_active = False
                self.vad_end_sent = True
                self.vad_start_sent = False
                speech_end_ms = vad_result_end.get("end_time") if self.vad else buffer.get_time_ms(current_offset)
                vad_events.append({
                    'type': 'end',
                    'timestamp_ms': int(speech_end_ms),
                    'confidence': vad_result_end.get('confidence', 0.0)
                })
                logger.info(f"Speech end event queued at {speech_end_ms}ms")

        audio_to_process = buffer.get_unprocessed_audio()
        if len(audio_to_process) == 0:
            logger.debug("No new audio to process")
            return None, None, "", vad_events

        if self.vad_enable and speech_start_ms is not None:
            start_sample = int((speech_start_ms / 1000) * self.SAMPLING_RATE)
            if speech_end_ms is not None:
                end_sample = min(int((speech_end_ms / 1000) * self.SAMPLING_RATE), len(audio_to_process))
            else:
                end_sample = len(audio_to_process)
            audio_to_process = audio_to_process[start_sample:end_sample]

        current_duration_ms = (len(audio_to_process) / self.SAMPLING_RATE) * 1000
        current_duration = current_duration_ms / 1000
        if current_duration < self.min_context_needed and not is_final:
            buffer.is_processing = False
            return None, None, "", vad_events

        try:
            prompt, context = self.construct_prompt()
            asr_results = self.asr.transcribe_to_segment(
                audio_to_process,
                init_prompt=prompt,
                language=self.vad_config.language if self.vad_enable else None,
            )
            timestamped_words = self.asr.segment_to_timestamped_words(asr_results)
            buffer_start_time = ((buffer.start_offset + buffer.last_processed_sample) / self.SAMPLING_RATE)
            self.transcript_buffer.insert(timestamped_words, buffer_start_time)

            if is_final:
                confirmed_transcriptions = self.transcript_buffer.set_final()
            else:
                confirmed_transcriptions = self.transcript_buffer.flush()

            self.committed_transcriptions.extend(confirmed_transcriptions)

            if not is_final:
                if self.trimming_mode == "sentence" and confirmed_transcriptions:
                    if current_duration > self.trimming_duration:
                        self.trim_to_last_completed_sentence()
                elif self.trimming_mode == "segment":
                    if current_duration > self.trimming_duration:
                        self.trim_to_last_completed_segment(asr_results)

            buffer.mark_processing_complete(is_final)

            start_time, end_time, transcription = self.format_output(confirmed_transcriptions)
            return start_time, end_time, transcription, vad_events

        except Exception as e:
            logger.error(f"Error in process_audio: {str(e)}")
            buffer.is_processing = False
            return None, None, "", vad_events

    def initialize_vad(self, vad_config: TurnDetectionConfig):
        logger.info("Initializing VAD processor")
        self.vad = VADProcessor(vad_config)

    def update_vad_config(self, vad_config: TurnDetectionConfig):
        """
        Update the VAD configuration and reset buffers if the mode changes.
        """
        new_vad_enable = vad_config.create_response if vad_config else False

        # Check if the VAD mode is changing
        if new_vad_enable != self.vad_enable:
            self.reset_buffers_for_mode_change(new_vad_enable)

        # Update the VAD configuration
        self.vad_config = vad_config if vad_config else TurnDetectionConfig()

        if new_vad_enable:
            # Initialize VAD and the VAD audio buffer if not already initialized
            if self.vad_audio_buffer is None:
                self.vad_audio_buffer = AudioBufferWithTiming(sample_rate=self.SAMPLING_RATE)
            self.initialize_vad(self.vad_config)
            self.vad_enable = True
        else:
            self.vad = None
            self.vad_enable = False

        # Reset VAD state tracking
        self.vad_speech_active = False
        self.vad_start_sent = False
        self.vad_end_sent = False


    def reset_state(self, initial_offset=0):
        """Resets the internal state of the processor."""
        self.audio_buffer.reset()
        if self.vad_enable:
            self.vad_audio_buffer.reset()  # Clear the VAD-specific buffer

        self.transcript_buffer = HypothesisBuffer()
        self.transcript_buffer.last_committed_time = initial_offset * 1000
        self.committed_transcriptions = []
        # Reset VAD state tracking
        self.vad_speech_active = False
        self.vad_start_sent = False
        self.vad_end_sent = False
        if self.vad:
            self.vad.reset()


    def save_debug_audio(self, audio_data: np.ndarray, suffix: str = ""):
        """
        Saves the current audio buffer to a WAV file for debugging.

        Args:
            audio_data: Audio data to save
            suffix: Optional suffix to add to the filename
        """
        if self.debug_audio_dir:
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            filename = f"audio_buffer_{timestamp}_{self.buffer_count}{suffix}.wav"
            filepath = os.path.join(self.debug_audio_dir, filename)

            try:
                sf.write(filepath, audio_data, self.SAMPLING_RATE)
                logger.info(f"Saved debug audio to {filepath}")
            except Exception as e:
                logger.error(f"Error saving debug audio: {str(e)}")

            self.buffer_count += 1


    def trim_buffer_at(self, timestamp):
        """
        Trims the audio buffer and transcription buffer at a specified timestamp.

        Args:
            timestamp: The time (in seconds) to trim the buffers up to.
        """
        self.transcript_buffer.pop_committed(timestamp)
        cut_samples = int((timestamp - (self.audio_buffer.start_offset / self.SAMPLING_RATE)) * self.SAMPLING_RATE)
        self.audio_buffer.clear_until(cut_samples)

    # Rest of the methods remain the same as they don't directly interact with the buffer
    def transcribe(
        self,
        audio: Union[str, BinaryIO, np.ndarray],
        init_prompt: str = "",
        temperature: float = 0.0,
        language: Optional[LanguageType] = None,
        include_segments: bool = False,
        batch: bool = False,
    ) -> TranscriptionResponse:

        return self.asr.transcribe(
            audio,
            init_prompt=init_prompt,
            temperature=temperature,
            language=language,
            include_segments=include_segments,
            batch=batch
        )

    async def transcribe_async(
        self, audio: Union[str, BinaryIO, np.ndarray],
        init_prompt: str = "",
        temperature: float = 0.0,
        language: str = None,
        include_segments: bool = False,
        batch: bool = False,
    ) -> TranscriptionResponse:
        loop = asyncio.get_event_loop()
        with ThreadPoolExecutor() as pool:
            return await loop.run_in_executor(
                pool,
                self.transcribe,
                audio, init_prompt, temperature, language, include_segments, batch
            )

    def construct_prompt(self):
        """
        Constructs a prompt from previously confirmed transcriptions for context during ASR.

        Returns:
            Tuple (prompt, context):
                - prompt: A 200-character string of previous context for the ASR model.
                - context: The remaining transcriptions within the current audio buffer for logging/debugging.
        """
        # Find the first committed transcription outside the current buffer
        start_index = max(0, len(self.committed_transcriptions) - 1)
        while start_index > 0 and self.committed_transcriptions[start_index - 1][1] > self.audio_buffer.get_time_ms(self.audio_buffer.last_processed_sample):
            start_index -= 1

        past_context = self.committed_transcriptions[:start_index]
        past_texts = [text for _, _, text in past_context]

        # Build prompt (up to 200 characters)
        prompt, current_length = [], 0
        while past_texts and current_length < 200:
            last_text = past_texts.pop()
            current_length += len(last_text) + 1
            prompt.append(last_text)

        remaining_transcriptions = self.committed_transcriptions[start_index:]
        return self.asr.sep.join(reversed(prompt)), self.asr.sep.join(text for _, _, text in remaining_transcriptions)

    def trim_to_last_completed_sentence(self):
        """
        Trims the audio buffer and committed transcriptions up to the end of the last completed sentence.
        """
        if not self.committed_transcriptions:
            return

        sentences = self.segment_transcriptions_into_sentences(self.committed_transcriptions)
        if len(sentences) < 2:
            return  # Not enough sentences to trim

        # Keep the last two sentences; trim up to the second-last one
        trim_timestamp = sentences[-2][1]
        self.trim_buffer_at(trim_timestamp)

    def trim_to_last_completed_segment(self, transcription_results):
        """
        Trims the audio buffer to the last completed audio segment.

        Args:
            transcription_results: Transcription results from the ASR system.
        """
        if not self.committed_transcriptions:
            return

        segment_end_times = self.asr.segments_end_ts(transcription_results)
        if len(segment_end_times) > 1:
            trim_timestamp = segment_end_times[-2] + (self.audio_buffer.start_offset / self.SAMPLING_RATE)
            if trim_timestamp <= self.committed_transcriptions[-1][1]:
                self.trim_buffer_at(trim_timestamp)

    def segment_transcriptions_into_sentences(self, words):
        """
        Segments committed transcriptions into sentences using the tokenizer.

        Args:
            words: List of transcribed words with timestamps.

        Returns:
            List of tuples [(start_time, end_time, sentence), ...].
        """
        if not self.tokenizer:
            raise ValueError("Tokenizer is required for sentence segmentation.")

        complete_words = list(words)
        text = " ".join(word[2] for word in complete_words)
        sentences = self.tokenizer.split(text)

        segmented = []
        for sentence in sentences:
            start, end = None, None
            while complete_words:
                word_start, word_end, word = complete_words.pop(0)
                if start is None and sentence.startswith(word):
                    start = word_start
                if sentence.strip() == word:
                    end = word_end
                    segmented.append((start, end, sentence.strip()))
                    break
                sentence = sentence[len(word):].strip()
        return segmented

    def format_output(self, sentences):
        """
        Formats the confirmed sentences for output.

        Args:
            sentences: List of sentences [(start_time, end_time, sentence), ...].

        Returns:
            A tuple (start_time, end_time, combined_text) or (None, None, "") if empty.
        """
        if not sentences:
            return None, None, ""

        start_time = sentences[0][0]
        end_time = sentences[-1][1]
        combined_text = self.asr.sep.join(sentence[2] for sentence in sentences)
        return start_time, end_time, combined_text


# Copyright (c) 2022 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Credits
    This code is modified from https://github.com/GitYCC/g2pW
"""
from typing import Dict
from typing import List
from typing import Tuple

import numpy as np

from .utils import tokenize_and_map

ANCHOR_CHAR = '▁'


def prepare_onnx_input(tokenizer,
                       labels: List[str],
                       char2phonemes: Dict[str, List[int]],
                       chars: List[str],
                       texts: List[str],
                       query_ids: List[int],
                       use_mask: bool=False,
                       window_size: int=None,
                       max_len: int=512) -> Dict[str, np.array]:
    if window_size is not None:
        truncated_texts, truncated_query_ids = _truncate_texts(
            window_size=window_size, texts=texts, query_ids=query_ids)
    input_ids = []
    token_type_ids = []
    attention_masks = []
    phoneme_masks = []
    char_ids = []
    position_ids = []

    for idx in range(len(texts)):
        text = (truncated_texts if window_size else texts)[idx].lower()
        query_id = (truncated_query_ids if window_size else query_ids)[idx]

        try:
            tokens, text2token, token2text = tokenize_and_map(
                tokenizer=tokenizer, text=text)
        except Exception:
            print(f'warning: text "{text}" is invalid')
            return {}

        text, query_id, tokens, text2token, token2text = _truncate(
            max_len=max_len,
            text=text,
            query_id=query_id,
            tokens=tokens,
            text2token=text2token,
            token2text=token2text)

        processed_tokens = ['[CLS]'] + tokens + ['[SEP]']

        input_id = list(
            np.array(tokenizer.convert_tokens_to_ids(processed_tokens)))
        token_type_id = list(np.zeros((len(processed_tokens), ), dtype=int))
        attention_mask = list(np.ones((len(processed_tokens), ), dtype=int))

        query_char = text[query_id]
        phoneme_mask = [1 if i in char2phonemes[query_char] else 0 for i in range(len(labels))] \
            if use_mask else [1] * len(labels)
        char_id = chars.index(query_char)
        position_id = text2token[
            query_id] + 1  # [CLS] token locate at first place

        input_ids.append(input_id)
        token_type_ids.append(token_type_id)
        attention_masks.append(attention_mask)
        phoneme_masks.append(phoneme_mask)
        char_ids.append(char_id)
        position_ids.append(position_id)

    outputs = {
        'input_ids': np.array(input_ids).astype(np.int64),
        'token_type_ids': np.array(token_type_ids).astype(np.int64),
        'attention_masks': np.array(attention_masks).astype(np.int64),
        'phoneme_masks': np.array(phoneme_masks).astype(np.float32),
        'char_ids': np.array(char_ids).astype(np.int64),
        'position_ids': np.array(position_ids).astype(np.int64),
    }
    return outputs


def _truncate_texts(window_size: int, texts: List[str],
                    query_ids: List[int]) -> Tuple[List[str], List[int]]:
    truncated_texts = []
    truncated_query_ids = []
    for text, query_id in zip(texts, query_ids):
        start = max(0, query_id - window_size // 2)
        end = min(len(text), query_id + window_size // 2)
        truncated_text = text[start:end]
        truncated_texts.append(truncated_text)

        truncated_query_id = query_id - start
        truncated_query_ids.append(truncated_query_id)
    return truncated_texts, truncated_query_ids


def _truncate(max_len: int,
              text: str,
              query_id: int,
              tokens: List[str],
              text2token: List[int],
              token2text: List[Tuple[int]]):
    truncate_len = max_len - 2
    if len(tokens) <= truncate_len:
        return (text, query_id, tokens, text2token, token2text)

    token_position = text2token[query_id]

    token_start = token_position - truncate_len // 2
    token_end = token_start + truncate_len
    font_exceed_dist = -token_start
    back_exceed_dist = token_end - len(tokens)
    if font_exceed_dist > 0:
        token_start += font_exceed_dist
        token_end += font_exceed_dist
    elif back_exceed_dist > 0:
        token_start -= back_exceed_dist
        token_end -= back_exceed_dist

    start = token2text[token_start][0]
    end = token2text[token_end - 1][1]

    return (text[start:end], query_id - start, tokens[token_start:token_end], [
        i - token_start if i is not None else None
        for i in text2token[start:end]
    ], [(s - start, e - start) for s, e in token2text[token_start:token_end]])


def get_phoneme_labels(polyphonic_chars: List[List[str]]
                       ) -> Tuple[List[str], Dict[str, List[int]]]:
    labels = sorted(list(set([phoneme for char, phoneme in polyphonic_chars])))
    char2phonemes = {}
    for char, phoneme in polyphonic_chars:
        if char not in char2phonemes:
            char2phonemes[char] = []
        char2phonemes[char].append(labels.index(phoneme))
    return labels, char2phonemes


def get_char_phoneme_labels(polyphonic_chars: List[List[str]]
                            ) -> Tuple[List[str], Dict[str, List[int]]]:
    labels = sorted(
        list(set([f'{char} {phoneme}' for char, phoneme in polyphonic_chars])))
    char2phonemes = {}
    for char, phoneme in polyphonic_chars:
        if char not in char2phonemes:
            char2phonemes[char] = []
        char2phonemes[char].append(labels.index(f'{char} {phoneme}'))
    return labels, char2phonemes



import time
import sys
import httpx
from fastapi import FastAPI, HTTPException, Request, APIRouter
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from starlette.types import ASGIApp, Receive, Scope, Send
from collections import defaultdict
from typing import Union
from gallama.data_classes.data_class import ModelSpec
from gallama.data_classes import  ModelInstanceInfo,  ModelInfo, AgentWithThinking
from gallama.server_engine import handle_mixture_of_agent_request, create_options_response
from gallama.utils import parse_request_body
from typing import List, Dict
from gallama.config import ConfigManager
from gallama.server_engine import forward_request
import shutil
import asyncio
import uvicorn
import argparse
from gallama.utils.utils import get_package_file_path
from logging import DEBUG, INFO
import os
import json
import base64
from gallama.server_routes import (
    server_management_router,
    realtime_router,
    periodic_health_check,
    stop_model_instance
)
from gallama.server_engine import log_model_status
from gallama.dependencies_server import get_server_manager, get_server_logger, start_log_receiver, DEFAULT_ZMQ_URL


server_logger = get_server_logger()

router = APIRouter()
router.include_router(server_management_router)
router.include_router(realtime_router)


manager_app = FastAPI()
manager_app.include_router(router)





@manager_app.middleware("http")
@manager_app.middleware("https")
async def log_requests(request: Request, call_next):
    try:
        if request.method in ("POST", "PUT", "PATCH"):  # Methods that typically have a body
            content_type = request.headers.get("Content-Type", "")
            if "multipart/form-data" in content_type or "application/octet-stream" in content_type:
                server_logger.debug(f"API Request:\nMethod: {request.method}\nURL: {request.url}\n[Binary content omitted]")
            else:
                request_content = await request.body()
                if request_content:
                    try:
                        request_content = json.dumps(json.loads(request_content.decode("utf-8")), indent=2)
                    except (UnicodeDecodeError, json.JSONDecodeError):
                        request_content = "[Non-JSON or binary content omitted]"
                    server_logger.info(f"API Request:\nMethod: {request.method}\nURL: {request.url}\n{request_content}")

        response = await call_next(request)
    except RequestValidationError as e:
        server_logger.debug(f"Validation error:\n{e}")
        response = JSONResponse(status_code=422, content={"detail": "Validation error"})

    return response


class WebSocketLoggingMiddleware:
    def __init__(self, app: ASGIApp):
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        if scope["type"] == "websocket":
            await self.log_websocket(scope, receive, send)
        else:
            await self.app(scope, receive, send)

    async def log_websocket(self, scope: Scope, receive: Receive, send: Send):
        # Log detailed connection attempt information
        client = scope.get('client')
        server_logger.info(f"WebSocket connection attempt from {client}")

        # Log query parameters
        query_string = scope.get('query_string', b'').decode('utf-8')
        server_logger.debug(f"Query parameters: {query_string}")

        # Convert headers to a more readable format and log them
        headers = dict(scope.get('headers', []))
        formatted_headers = {
            k.decode('utf-8'): v.decode('utf-8')
            for k, v in headers.items()
        }
        server_logger.debug("WebSocket headers:")
        for key, value in formatted_headers.items():
            server_logger.debug(f"  {key}: {value}")

        # Log protocols specifically
        protocols = formatted_headers.get('sec-websocket-protocol', '').split(', ')
        server_logger.debug(f"Requested protocols: {protocols}")

        async def receive_with_logging():
            message = await receive()
            server_logger.debug(f"Received message type: {message['type']}")

            if message["type"] == "websocket.connect":
                server_logger.info("WebSocket client connecting")
                # Log any authentication info (safely)
                auth_header = formatted_headers.get('authorization', '')
                if auth_header:
                    server_logger.debug("Authorization header present")

            elif message["type"] == "websocket.disconnect":
                server_logger.info(f"WebSocket client disconnecting from {client}")

            elif message["type"] == "websocket.receive":
                try:
                    if "text" in message:
                        content = message["text"]
                        try:
                            parsed_content = json.loads(content)
                            server_logger.info(
                                "WebSocket received text:\n" +
                                json.dumps(parsed_content, indent=2)
                            )
                        except json.JSONDecodeError:
                            server_logger.info(f"WebSocket received raw text: {content}")
                    elif "bytes" in message:
                        server_logger.info(f"WebSocket received binary data of size: {len(message['bytes'])} bytes")
                except Exception as e:
                    server_logger.error(f"Error logging WebSocket message: {str(e)}")
                    server_logger.debug(f"Raw message content: {message}")
            return message

        async def send_with_logging(message):
            server_logger.debug(f"Sending message type: {message['type']}")

            if message["type"] == "websocket.accept":
                selected_protocol = message.get('subprotocol')
                server_logger.info(f"WebSocket connection accepted with protocol: {selected_protocol}")

            elif message["type"] == "websocket.close":
                code = message.get('code', 1000)
                reason = message.get('reason', '')
                server_logger.info(f"WebSocket connection closed with code {code}: {reason}")

            elif message["type"] == "websocket.send":
                try:
                    if "text" in message:
                        content = message["text"]
                        try:
                            parsed_content = json.loads(content)
                            server_logger.info(
                                "WebSocket sent text:\n" +
                                json.dumps(parsed_content, indent=2)
                            )
                        except json.JSONDecodeError:
                            server_logger.info(f"WebSocket sent raw text: {content}")
                    elif "bytes" in message:
                        server_logger.info(f"WebSocket sent binary data of size: {len(message['bytes'])} bytes")
                except Exception as e:
                    server_logger.error(f"Error logging WebSocket message: {str(e)}")
                    server_logger.debug(f"Raw message content: {message}")

            # Log any error responses
            if message.get('status', 200) >= 400:
                server_logger.error(f"WebSocket error response: {message}")

            await send(message)

        try:
            await self.app(scope, receive_with_logging, send_with_logging)
        except Exception as e:
            server_logger.error(f"WebSocket connection error: {str(e)}")
            server_logger.debug(f"Error details:", exc_info=True)

# Add WebSocket middleware
# manager_app.add_middleware(WebSocketLoggingMiddleware)

# Add CORS middleware
manager_app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # Allows all origins
    allow_credentials=True,
    allow_methods=["*"],  # Allows all methods
    allow_headers=["*"],  # Allows all headers
)


# models: Dict[str, ModelInfo] = {}
# model_load_queue = asyncio.Queue()
# loading_lock = asyncio.Lock()
# active_requests_lock = asyncio.Lock()
# task_status = {}
config_manager = ConfigManager()

# Define a variable for the starting port
START_PORT = 8001

# Increase the timeout for long-running tasks (adjust as needed)
TIMEOUT = 300  # 5 minutes

strict_mode = False

# List of endpoints to exclude from API gateway redirection
EXCLUDED_ENDPOINTS = [
    "/add_model", "/remove_model", "/list_models", "/list_available_models",
    "/v1/add_model", "/v1/remove_model", "/v1/list_models", "/v1/list_available_models",
    "/list_models", "/v1/list_models"
    "/v1/remove_model_by_port","/remove_model_by_port",
    "/v1/models", "/task_status"
]
EMBEDDING_SUBPATHS = []     # No overwriting at the moment
# EMBEDDING_SUBPATHS = [
#     {
#         "original": "/v1/embeddings",
#         "replacement": "/v1/embeddings"
#     },
#     # Add more paths here as needed, e.g.:
#     # {"original": "/v2/some_path", "replacement": "/new_path"}
# ]

# Dictionary to keep track of active requests per instance
active_requests: Dict[int, int] = defaultdict(int)

# Add a semaphore for concurrency control
MAX_CONCURRENT_REQUESTS = 100  # Adjust this value based on your system's capacity
request_semaphore = asyncio.Semaphore(MAX_CONCURRENT_REQUESTS)


# # set up logging
# def start_log_receiver(zmq_url):
#     context = zmq.Context()
#     socket = context.socket(zmq.PULL)
#     socket.bind("tcp://*:5555")
#
#     # Initialize the server_logger for the receiver
#     receiver_logger = get_logger(name="log_receiver", to_console=True, to_file=False, to_zmq=False)
#
#     receiver_logger.info(f"Log receiver started on {zmq_url}")
#
#     def receive_logs():
#         while True:
#             try:
#                 message = socket.recv_json()
#                 log_level = getattr(logging, message['level'].upper(), logging.INFO)
#
#                 receiver_logger.log(
#                     level=log_level,
#                     msg=f"{message['model']}:{message['port']} | {message['log']}"
#                 )
#             except zmq.Again:
#                 # No message available, sleep for a short time
#                 time.sleep(0.1)
#             except zmq.ZMQError as e:
#                 receiver_logger.error(f"ZMQ Error in log receiver: {e}")
#             except json.JSONDecodeError as e:
#                 receiver_logger.error(f"JSON Decode Error in log receiver: {e}")
#             except Exception as e:
#                 receiver_logger.error(f"Unexpected error in log receiver: {e}")
#                 receiver_logger.error(traceback.format_exc())
#
#     # Start the receiver in a separate thread
#     receiver_thread = threading.Thread(target=receive_logs, daemon=True)
#     receiver_thread.start()
#
#     return receiver_thread
#
#
#
# # Start the log receiver in a separate thread
# DEFAULT_ZMQ_URL = "tcp://127.0.0.1:5555"  # Using 5559 as a standard port for logging
#
#
# # Initialize the logger for the manager
# # Set to_console=True and to_zmq=False to avoid duplication
# logger = get_logger(name="manager", to_console=True, to_zmq=False)



async def model_loader():
    server_manager = get_server_manager()

    while True:
        server_logger.info("Model loader waiting for next model in queue")
        model = await server_manager.model_load_queue.get()
        server_logger.info(f"Model loader retrieved model from queue: {model}")
        async with server_manager.loading_lock:
            try:
                server_logger.info(f"Starting to load model: {model}")
                server_logger.info(f"Current queue size: {server_manager.model_load_queue.qsize()}")
                await run_model_with_timeout(model)
            except Exception as e:
                server_logger.exception(f"Error loading model {model} instance: {str(e)}")
            finally:
                server_logger.info(f"Finished processing model: {model}")
                server_manager.model_load_queue.task_done()
                server_logger.info(f"Remaining queue size: {server_manager.model_load_queue.qsize()}")

        server_logger.info("Model loader finished processing, looping back")
        await asyncio.sleep(1)


async def wait_for_model_ready(port, timeout=300):  # 5 minutes timeout
    start_time = time.time()
    while time.time() - start_time < timeout:
        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(f"http://localhost:{port}/health", timeout=5.0)
                if response.status_code == 200:
                    return True
        except httpx.RequestError:
            pass  # The server_engine might not be up yet, so we'll just continue waiting
        await asyncio.sleep(1)
    return False


async def run_model(model_spec: ModelSpec):
    server_manager = get_server_manager()

    try:
        # add the entry for the dictionary (where key is the model name)
        if model_spec.model_name not in server_manager.models:
            server_manager.models[model_spec.model_name] = ModelInfo(instances=[])

        # Find the next available port
        port = START_PORT
        while any(instance.port == port for model_info in server_manager.models.values() for instance in model_info.instances):
            port += 1

        server_logger.info(f"Attempting to start model {model_spec.model_name} on port {port}")

        model_config = config_manager.configs.get(model_spec.model_name)
        backend = model_spec.backend or model_config.get('backend')

        try:
            # Serialize the ModelSpec to JSON and encode to base64
            model_json = model_spec.model_dump_json()
            model_b64 = base64.b64encode(model_json.encode('utf-8')).decode('utf-8')


            # Use the function
            app_path = get_package_file_path('app.py')
            server_logger.info(f"Using app path: {app_path}")

            # Determine the correct Python executable
            python_exec = shutil.which("python3") or shutil.which("python")

            if backend == "exllama":
                # model_cli_args = model.to_arg_string()
                # server_logger.debug(f"model cli: {model_cli_args}")
                process = await asyncio.create_subprocess_exec(
                    python_exec, app_path, "--model-spec", model_b64, "--detached", "--port", str(port),
                    stdout=asyncio.subprocess.DEVNULL,
                    # stderr=asyncio.subprocess.DEVNULL,
                )
            else:
                # other than exllama, we will use env setting to set visible GPUs
                # CUDA_VISIBLE_DEVICES constraints before launching fastapi

                # Create a copy of the current environment
                env = os.environ.copy()

                # Set CUDA_VISIBLE_DEVICES
                env['CUDA_VISIBLE_DEVICES'] = model_spec.get_visible_gpu_indices()

                # model_cli_args = model.to_arg_string()
                # server_logger.debug(f"model cli: {model_cli_args}")
                process = await asyncio.create_subprocess_exec(
                    python_exec, app_path, "--model-spec", model_b64, "--detached", "--port", str(port),
                    stdout=asyncio.subprocess.DEVNULL,
                    # stderr=asyncio.subprocess.DEVNULL,
                    env=env  # Pass the modified environment to the subprocess
                )

        except Exception as e:
            server_logger.error(f"Failed to create subprocess for model {model_spec.model_id} on port {port}: {str(e)}")
            return

        # Wait for the model to become ready
        if await wait_for_model_ready(port):
            server_logger.info(f"Model {model_spec.model_name} on port {port} is ready")
            instance_info = ModelInstanceInfo(
                port=port,
                pid=process.pid,
                status="running",
                model_name=model_spec.model_name,
                model_type=ModelSpec.get_model_type_from_backend(backend),
                strict=model_spec.strict
            )
            server_manager.models[model_spec.model_name].instances.append(instance_info)
        else:
            server_logger.error(f"Timeout waiting for model {model_spec.model_name} on port {port} to become ready")
            await stop_model_instance(model_spec.model_name, port)
            return

        server_logger.info(f"Model {model_spec.model_id} instance on port {port} is fully loaded and ready")
        log_model_status(server_manager.models, custom_logger=server_logger)  # Log status after successfully loading a model

        # Instead of entering an infinite loop, we'll exit the function here
        return

    except Exception as e:
        server_logger.exception(f"Error running model {model_spec.model_name} instance on port {port}: {str(e)}")
        await stop_model_instance(model_spec.model_name, port)
    finally:
        await cleanup_after_model_load(model_spec.model_name)
        server_logger.info(f"Exiting run_model for {model_spec.model_name} on port {port}")


async def cleanup_after_model_load(model: str):
    # Perform any necessary cleanup here
    server_logger.info(f"Performing cleanup after loading model: {model}")
    # For example, you might want to close any open connections or release resources
    # This is a placeholder - add specific cleanup tasks as needed


async def run_model_with_timeout(model: str, timeout: int = 600):  # 10 minutes timeout
    server_manager = get_server_manager()

    try:
        await asyncio.wait_for(run_model(model), timeout=timeout)
    except asyncio.TimeoutError:
        server_logger.error(f"Timeout occurred while loading model: {model}")
        # Perform cleanup or error handling here
        await stop_model_instance(model, server_manager.models[model].instances[-1].port if model in server_manager.models and server_manager.models[model].instances else None)




async def get_model_from_body(request: Request) -> str:
    try:
        body, is_multipart = await parse_request_body(request)
        return body.get("model", "")
    except Exception as e:     # when request doesnt come with model e.g. audio transcription
        server_logger.debug(f"Can not get body from request {e}. Hence continue with raw body")
        return ""



async def get_instance_for_model(model: str):
    server_manager = get_server_manager()

    running_instances = [inst for inst in server_manager.models[model].instances if inst.status == "running"]
    if not running_instances:
        raise HTTPException(status_code=503, detail=f"No running instances available for model: {model}")

    return min(running_instances, key=lambda inst: active_requests[inst.port])


async def forward_to_multiple_agents(request: Request, agent_list: List[Union[str, AgentWithThinking]], modified_body: str, modified_headers: str):
    tasks = []
    for agent in agent_list:
        if isinstance(agent, str):
            instance = await get_instance_for_model(agent)
            tasks.append(forward_request(request, instance, modified_body, modified_headers))
        elif isinstance(agent, AgentWithThinking):
            instance = await get_instance_for_model(agent.model)
            if agent.thinking_template:
                # Modify the request to include the thinking in the thinking_template field
                agent_body = json.loads(modified_body)
                agent_body["thinking_template"] = agent.thinking_template
                agent_modified_body = json.dumps(agent_body).encode()
                agent_modified_headers = dict(modified_headers)
                agent_modified_headers["content-length"] = str(len(agent_modified_body))
                tasks.append(forward_request(request, instance, agent_modified_body, agent_modified_headers))
            else:
                tasks.append(forward_request(request, instance, modified_body, modified_headers))
    return await asyncio.gather(*tasks)



async def load_balanced_router(request: Request, path: str):
    server_manager = get_server_manager()

    body_json, is_multipart = await parse_request_body(request)

    if request.url.path in EXCLUDED_ENDPOINTS:
        server_logger.info(body_json)
        return await request.app.router.get_route_handler()(request)

    if request.method == "OPTIONS":
        return create_options_response(dict(request.headers))

    if isinstance(body_json, dict) and body_json.get("mixture_of_agents", False):
        return await handle_mixture_of_agent_request(request, body_json, server_manager.models, active_requests)

    model = await get_model_from_body(request)

    async with request_semaphore:
        is_embedding = any(subpath["original"] in path for subpath in EMBEDDING_SUBPATHS)

        if not model and strict_mode:
            raise HTTPException(status_code=400, detail="Model must be specified in strict mode")

        available_instances = []

        if strict_mode:
            # In strict mode, the model must be specified and found
            if model not in server_manager.models:
                raise HTTPException(status_code=404, detail="Specified model not found")
            available_instances = [inst for inst in server_manager.models[model].instances if inst.status == "running"]
        else:
            # Try to find a matching model first (if model is specified)
            if model and model in server_manager.models:
                available_instances = [inst for inst in server_manager.models[model].instances if inst.status == "running"]

            # If no matching model or no instances found, pick any running instance of the correct type that is not strict
            if not available_instances:
                # Determine the model type based on the request path
                if "/audio/transcriptions" in path:
                    model_type = "stt"
                elif "embeddings" in path:
                    model_type = "embedding"
                elif "/audio/speech" in path:
                    model_type = "tts"
                else:
                    model_type = "llm"

                for model_info in server_manager.models.values():
                    for inst in model_info.instances:
                        if inst.status == "running" and inst.model_type == model_type and not inst.strict:
                            available_instances.append(inst)

        if not available_instances:
            raise HTTPException(status_code=503, detail=f"No suitable running instances with requested model '{model}'")

        # Select the instance with the least active requests
        instance = min(available_instances, key=lambda inst: active_requests[inst.port])

        # Increment the active request count for the selected instance
        async with server_manager.active_requests_lock:
            active_requests[instance.port] += 1

        try:
            server_logger.info(f"active_requests: {str(active_requests)}")
            server_logger.info(f"Request routed to model {instance.model_name} instance at port {instance.port}")

            # Forward the request
            response = await forward_request(request, instance)

            return response
        finally:
            # Decrement the active request count after the request is completed
            async with server_manager.active_requests_lock:
                active_requests[instance.port] -= 1


@manager_app.api_route("/{path:path}", methods=["GET", "POST", "PUT", "DELETE", "PATCH", "OPTIONS"])
async def router(request: Request, path: str):
    return await load_balanced_router(request, path)


def shutdown(signal, frame):
    server_manager = get_server_manager()

    server_logger.info("Shutting down all models...")
    for model in list(server_manager.models.keys()):
        for instance in server_manager.models[model].instances:
            stop_model_instance(model, instance.port)
    server_logger.info("All models have been shut down. Exiting.")
    sys.exit(0)


async def start_server(port=8000):
    config = uvicorn.Config(manager_app, host="0.0.0.0", port=port, log_level="debug")
    server = uvicorn.Server(config)
    await server.serve()


async def main(model_list=None, port=8000, strict_mode=False):
    server_manager = get_server_manager()

    receiver_thread = start_log_receiver(DEFAULT_ZMQ_URL)
    server_logger.info("Starting main function")
    server_logger.info(f"Strict mode: {'enabled' if strict_mode else 'disabled'}")

    model_loader_task = asyncio.create_task(model_loader())
    server_logger.info("Created model_loader task")

    # Load initial models
    if model_list:
        server_logger.info(f"Loading initial models: {[model.model_name for model in model_list]}")
        for model in model_list:
            await server_manager.model_load_queue.put(model)

        # Wait for all initial models to load
        while not server_manager.model_load_queue.empty():
            await asyncio.sleep(1)

        server_logger.info("All initial models have been queued for loading")

    # Start periodic health checks
    health_check_task = asyncio.create_task(periodic_health_check())

    await start_server(port)

    # Ensure tasks are cancelled when the server_engine stops
    model_loader_task.cancel()
    health_check_task.cancel()
    try:
        await asyncio.gather(model_loader_task, health_check_task)
    except asyncio.CancelledError:
        server_logger.info("Tasks cancelled")


# this parser is different to app.py
def parse_dict(arg):
    """Parses a key=value string and returns a dictionary."""
    result = {}
    for pair in arg.split():
        key, value = pair.split('=')
        result[key] = value.strip("'")  # Strip single quotes here as well
    return result


def llama_picture():
    """ show a fun picture cause why not :) """
    # Load the image
    llama_art = """
        ,,__
        .. \\(oo)
           (__)
             ||----w |
             ||     ||
    """
    server_logger.info(llama_art)


def run_from_script(args):
    # opening llama picture cause why not
    llama_picture()

    global strict_mode

    if args.strict_mode:
        strict_mode = True

    # parse the cli input
    model_list = []
    if args.model_id:
        for item in args.model_id:
            model_spec = ModelSpec.from_dict(item)
            model_list.append(model_spec)
            server_logger.info(model_spec)

    # simple loading with model_name
    if args.model_name:
        model_spec = ModelSpec(model_name=args.model_name)
        model_list.append(model_spec)

    if model_list:
        server_logger.info("Initial models: " + str(model_list))

    # set server_logger level
    if args.verbose:
        server_logger.setLevel(DEBUG)
    else:
        server_logger.setLevel(INFO)


    server_logger.info("Parsed Arguments:" + str(args))

    asyncio.run(
        main(
            model_list=model_list,
            port=args.port,
            strict_mode=args.strict_mode
        )
    )


if __name__ == "__main__":

    server_logger.info("Script started")
    arg_parser = argparse.ArgumentParser(description="Launch multi model src instance")
    arg_parser.add_argument("--strict_mode", action="store_true", default=False,
                            help="Enable strict mode for routing non-embedding requests to matching model names")
    arg_parser.add_argument("-id", "--model_id", action='append', type=parse_dict, default=None,
                            help="Model configuration. Can be used multiple times for different models. "
                                 "Format: -id model_id=<model_name> [gpu=<vram1>,<vram2>,...] [gpu<n>=<vram>] [cache=<size>] [model_type=exllama|embedding]"
                                 "Examples:\n"
                                 "1. Specify all GPUs: -id model_id=/path/llama3-8b gpu=4,2,0,3 cache=8192\n"
                                 "2. Specify individual GPUs: -id model_id=/path/llama3-8b gpu0=8 gpu1=8 cache=8192\n"
                                 "3. Mix formats: -id model_id=gpt4 gpu=4,2 gpu3=3 cache=8192\n"
                                 "4. Embedding model: -id model_id=/path/gte-large-en-v1.5 gpu=0,0,0,3"
                                 "cache_size is defaulted to model context length if not specified"
                                 "cache_size can be more than context length, which will help model perform better in batched generation"
                                 "cache_size is not application to embedding model"
                                 "VRAM is specified in GB. Cache size is integer which is the context length to cache."
                                 "VRAM for embedding will simple set env parameter to allow infinity_embedding to view the specific GPU and can not enforce VRAM size restriction")
    arg_parser.add_argument('-v', "--verbose", action='store_true', help="Turn on more verbose logging")
    arg_parser.add_argument("--host", type=str, default="127.0.0.1", help="The host to bind to.")
    arg_parser.add_argument('-p', "--port", type=int, default=8000, help="The port to bind to.")

    # arg_parser.add_argument("--reload", action="store_true", help="Enable auto-reload.")

    args = arg_parser.parse_args()

    run_from_script(args)




# modified from https://github.com/yangdongchao/SoundStorm/blob/master/soundstorm/s1/AR/models/t2s_model.py
# reference: https://github.com/lifeiteng/vall-e
import math
from typing import List, Optional
import torch
from tqdm import tqdm

from AR.models.utils import make_pad_mask
from AR.models.utils import (
    topk_sampling,
    sample,
    logits_to_probs,
    multinomial_sample_one_no_sync,
    dpo_loss,
    make_reject_y,
    get_batch_logps
)
from AR.modules.embedding import SinePositionalEmbedding
from AR.modules.embedding import TokenEmbedding
from AR.modules.transformer import LayerNorm
from AR.modules.transformer import TransformerEncoder
from AR.modules.transformer import TransformerEncoderLayer
from torch import nn
from torch.nn import functional as F
from torchmetrics.classification import MulticlassAccuracy

default_config = {
    "embedding_dim": 512,
    "hidden_dim": 512,
    "num_head": 8,
    "num_layers": 12,
    "num_codebook": 8,
    "p_dropout": 0.0,
    "vocab_size": 1024 + 1,
    "phoneme_vocab_size": 512,
    "EOS": 1024,
}

# @torch.jit.script ## 使用的话首次推理会非常慢，而且推理速度不稳定
# Efficient implementation equivalent to the following:
def scaled_dot_product_attention(query:torch.Tensor, key:torch.Tensor, value:torch.Tensor, attn_mask:Optional[torch.Tensor]=None, scale:Optional[torch.Tensor]=None) -> torch.Tensor:
    B, H, L, S =query.size(0), query.size(1), query.size(-2), key.size(-2)
    if scale is None:
        scale_factor = torch.tensor(1 / math.sqrt(query.size(-1)))
    else:
        scale_factor = scale
    attn_bias = torch.zeros(B, H, L, S, dtype=query.dtype, device=query.device)

    if attn_mask is not None:
        if attn_mask.dtype == torch.bool:
            attn_bias.masked_fill_(attn_mask, float("-inf"))
        else:
            attn_bias += attn_mask
    attn_weight = query @ key.transpose(-2, -1) * scale_factor
    attn_weight += attn_bias
    attn_weight = torch.softmax(attn_weight, dim=-1)

    if attn_mask is not None:
        if attn_mask.dtype == torch.bool:
            attn_weight.masked_fill_(attn_mask, 0)
        else:
            attn_mask[attn_mask!=float("-inf")] =0
            attn_mask[attn_mask==float("-inf")] =1
            attn_weight.masked_fill_(attn_mask, 0)

    return attn_weight @ value

@torch.jit.script
class T2SMLP:
    def __init__(self, w1, b1, w2, b2):
        self.w1 = w1
        self.b1 = b1
        self.w2 = w2
        self.b2 = b2

    def forward(self, x):
        x = F.relu(F.linear(x, self.w1, self.b1))
        x = F.linear(x, self.w2, self.b2)
        return x


@torch.jit.script
class T2SBlock:
    def __init__(
            self,
            num_heads,
            hidden_dim: int,
            mlp: T2SMLP,
            qkv_w,
            qkv_b,
            out_w,
            out_b,
            norm_w1,
            norm_b1,
            norm_eps1,
            norm_w2,
            norm_b2,
            norm_eps2,
    ):
        self.num_heads = num_heads
        self.mlp = mlp
        self.hidden_dim: int = hidden_dim
        self.qkv_w = qkv_w
        self.qkv_b = qkv_b
        self.out_w = out_w
        self.out_b = out_b
        self.norm_w1 = norm_w1
        self.norm_b1 = norm_b1
        self.norm_eps1 = norm_eps1
        self.norm_w2 = norm_w2
        self.norm_b2 = norm_b2
        self.norm_eps2 = norm_eps2

        self.false = torch.tensor(False, dtype=torch.bool)

    @torch.jit.ignore
    def to_mask(self, x:torch.Tensor, padding_mask:Optional[torch.Tensor]):
        if padding_mask is None:
            return x
        
        if padding_mask.dtype == torch.bool:
            return x.masked_fill(padding_mask, 0)
        else:
            return x * padding_mask
        
    def process_prompt(self, x:torch.Tensor, attn_mask : torch.Tensor, padding_mask:Optional[torch.Tensor]=None, torch_sdpa:bool=True):

            
        q, k, v = F.linear(self.to_mask(x, padding_mask), self.qkv_w, self.qkv_b).chunk(3, dim=-1)

        batch_size = q.shape[0]
        q_len = q.shape[1]
        kv_len = k.shape[1]
        
        q = self.to_mask(q, padding_mask)
        k_cache = self.to_mask(k, padding_mask)
        v_cache = self.to_mask(v, padding_mask)

        q = q.view(batch_size, q_len, self.num_heads, -1).transpose(1, 2)
        k = k_cache.view(batch_size, kv_len, self.num_heads, -1).transpose(1, 2)
        v = v_cache.view(batch_size, kv_len, self.num_heads, -1).transpose(1, 2)

        if torch_sdpa:
            attn = F.scaled_dot_product_attention(q, k, v, ~attn_mask)
        else:
            attn = scaled_dot_product_attention(q, k, v, attn_mask)

        attn = attn.permute(2, 0, 1, 3).reshape(batch_size*q_len, self.hidden_dim)
        attn = attn.view(q_len, batch_size, self.hidden_dim).transpose(1, 0)
        attn = F.linear(self.to_mask(attn, padding_mask), self.out_w, self.out_b)

        if padding_mask is not None:
            for i in range(batch_size):
                # mask = padding_mask[i,:,0]
                if self.false.device!= padding_mask.device:
                    self.false = self.false.to(padding_mask.device)
                idx = torch.where(padding_mask[i,:,0]==self.false)[0]
                x_item = x[i,idx,:].unsqueeze(0)
                attn_item = attn[i,idx,:].unsqueeze(0)
                x_item = x_item + attn_item
                x_item = F.layer_norm(
                    x_item, [self.hidden_dim], self.norm_w1, self.norm_b1, self.norm_eps1
                )
                x_item = x_item + self.mlp.forward(x_item)
                x_item = F.layer_norm(
                    x_item,
                    [self.hidden_dim],
                    self.norm_w2,
                    self.norm_b2,
                    self.norm_eps2,
                )
                x[i,idx,:] = x_item.squeeze(0)
            x = self.to_mask(x, padding_mask)
        else:
            x = x + attn
            x = F.layer_norm(
                x, [self.hidden_dim], self.norm_w1, self.norm_b1, self.norm_eps1
            )
            x = x + self.mlp.forward(x)
            x = F.layer_norm(
                x,
                [self.hidden_dim],
                self.norm_w2,
                self.norm_b2,
                self.norm_eps2,
            )
        return x, k_cache, v_cache
    
    def decode_next_token(self, x:torch.Tensor, k_cache:torch.Tensor, v_cache:torch.Tensor, attn_mask:Optional[torch.Tensor]=None, torch_sdpa:bool=True):
        q, k, v = F.linear(x, self.qkv_w, self.qkv_b).chunk(3, dim=-1)

        k_cache = torch.cat([k_cache, k], dim=1)
        v_cache = torch.cat([v_cache, v], dim=1)
        
        batch_size = q.shape[0]
        q_len = q.shape[1]
        kv_len = k_cache.shape[1]

        q = q.view(batch_size, q_len, self.num_heads, -1).transpose(1, 2)
        k = k_cache.view(batch_size, kv_len, self.num_heads, -1).transpose(1, 2)
        v = v_cache.view(batch_size, kv_len, self.num_heads, -1).transpose(1, 2)


        if torch_sdpa:
            attn = F.scaled_dot_product_attention(q, k, v)
        else:
            attn = scaled_dot_product_attention(q, k, v, attn_mask)

        attn = attn.permute(2, 0, 1, 3).reshape(batch_size*q_len, self.hidden_dim)
        attn = attn.view(q_len, batch_size, self.hidden_dim).transpose(1, 0)
        attn = F.linear(attn, self.out_w, self.out_b)

        x = x + attn
        x = F.layer_norm(
            x, [self.hidden_dim], self.norm_w1, self.norm_b1, self.norm_eps1
        )
        x = x + self.mlp.forward(x)
        x = F.layer_norm(
            x,
            [self.hidden_dim],
            self.norm_w2,
            self.norm_b2,
            self.norm_eps2,
        )
        return x, k_cache, v_cache


@torch.jit.script
class T2STransformer:
    def __init__(self, num_blocks : int, blocks: List[T2SBlock]):
        self.num_blocks : int = num_blocks
        self.blocks = blocks

    def process_prompt(
        self, x:torch.Tensor, attn_mask : torch.Tensor,
        padding_mask : Optional[torch.Tensor]=None, 
        torch_sdpa:bool=True
        ):
        k_cache : List[torch.Tensor] = []
        v_cache : List[torch.Tensor] = []
        for i in range(self.num_blocks):
            x, k_cache_, v_cache_ = self.blocks[i].process_prompt(x, attn_mask, padding_mask, torch_sdpa)
            k_cache.append(k_cache_)
            v_cache.append(v_cache_)
        return x, k_cache, v_cache

    def decode_next_token(
        self, x:torch.Tensor, 
        k_cache: List[torch.Tensor], 
        v_cache: List[torch.Tensor], 
        attn_mask : Optional[torch.Tensor]=None,
        torch_sdpa:bool=True
    ):
        for i in range(self.num_blocks):
            x, k_cache[i], v_cache[i] = self.blocks[i].decode_next_token(x, k_cache[i], v_cache[i], attn_mask, torch_sdpa)
        return x, k_cache, v_cache


class Text2SemanticDecoder(nn.Module):
    def __init__(self, config, norm_first=False, top_k=3):
        super(Text2SemanticDecoder, self).__init__()
        self.model_dim = config["model"]["hidden_dim"]
        self.embedding_dim = config["model"]["embedding_dim"]
        self.num_head = config["model"]["head"]
        self.num_layers = config["model"]["n_layer"]
        self.norm_first = norm_first
        self.vocab_size = config["model"]["vocab_size"]
        self.phoneme_vocab_size = config["model"]["phoneme_vocab_size"]
        self.p_dropout = config["model"]["dropout"]
        self.EOS = config["model"]["EOS"]
        self.norm_first = norm_first
        assert self.EOS == self.vocab_size - 1
        # should be same as num of kmeans bin
        # assert self.EOS == 1024
        self.bert_proj = nn.Linear(1024, self.embedding_dim)
        self.ar_text_embedding = TokenEmbedding(
            self.embedding_dim, self.phoneme_vocab_size, self.p_dropout
        )
        self.ar_text_position = SinePositionalEmbedding(
            self.embedding_dim, dropout=0.1, scale=False, alpha=True
        )
        self.ar_audio_embedding = TokenEmbedding(
            self.embedding_dim, self.vocab_size, self.p_dropout
        )
        self.ar_audio_position = SinePositionalEmbedding(
            self.embedding_dim, dropout=0.1, scale=False, alpha=True
        )

        self.h = TransformerEncoder(
            TransformerEncoderLayer(
                d_model=self.model_dim,
                nhead=self.num_head,
                dim_feedforward=self.model_dim * 4,
                dropout=0.1,
                batch_first=True,
                norm_first=norm_first,
            ),
            num_layers=self.num_layers,
            norm=LayerNorm(self.model_dim) if norm_first else None,
        )

        self.ar_predict_layer = nn.Linear(self.model_dim, self.vocab_size, bias=False)
        self.loss_fct = nn.CrossEntropyLoss(reduction="sum")

        self.ar_accuracy_metric = MulticlassAccuracy(
            self.vocab_size,
            top_k=top_k,
            average="micro",
            multidim_average="global",
            ignore_index=self.EOS,
        )

        blocks = []

        for i in range(self.num_layers):
            layer = self.h.layers[i]
            t2smlp = T2SMLP(
                layer.linear1.weight,
                layer.linear1.bias,
                layer.linear2.weight,
                layer.linear2.bias
            )

            block = T2SBlock(
                self.num_head,
                self.model_dim,
                t2smlp,
                layer.self_attn.in_proj_weight,
                layer.self_attn.in_proj_bias,
                layer.self_attn.out_proj.weight,
                layer.self_attn.out_proj.bias,
                layer.norm1.weight,
                layer.norm1.bias,
                layer.norm1.eps,
                layer.norm2.weight,
                layer.norm2.bias,
                layer.norm2.eps
            )

            blocks.append(block)
        
        self.t2s_transformer = T2STransformer(self.num_layers, blocks)

    def make_input_data(self, x, x_lens, y, y_lens, bert_feature):
        x = self.ar_text_embedding(x)
        x = x + self.bert_proj(bert_feature.transpose(1, 2))
        x = self.ar_text_position(x)
        x_mask = make_pad_mask(x_lens)

        y_mask = make_pad_mask(y_lens)
        y_mask_int = y_mask.type(torch.int64)
        codes = y.type(torch.int64) * (1 - y_mask_int)

        # Training
        # AR Decoder
        y, targets = self.pad_y_eos(codes, y_mask_int, eos_id=self.EOS)
        x_len = x_lens.max()
        y_len = y_lens.max()
        y_emb = self.ar_audio_embedding(y)
        y_pos = self.ar_audio_position(y_emb)

        xy_padding_mask = torch.concat([x_mask, y_mask], dim=1)

        ar_xy_padding_mask = xy_padding_mask

        x_attn_mask = F.pad(
            torch.zeros((x_len, x_len), dtype=torch.bool, device=x.device),
            (0, y_len),
            value=True,
        )
        # x_attn_mask[:, x_len]=False
        y_attn_mask = F.pad(
            torch.triu(
                torch.ones(y_len, y_len, dtype=torch.bool, device=x.device),
                diagonal=1,
            ),
            (x_len, 0),
            value=False,
        )

        xy_attn_mask = torch.concat([x_attn_mask, y_attn_mask], dim=0)
        bsz, src_len = x.shape[0], x_len + y_len
        _xy_padding_mask = (
            ar_xy_padding_mask.view(bsz, 1, 1, src_len)
            .expand(-1, self.num_head, -1, -1)
            .reshape(bsz * self.num_head, 1, src_len)
        )
        xy_attn_mask = xy_attn_mask.logical_or(_xy_padding_mask)
        new_attn_mask = torch.zeros_like(xy_attn_mask, dtype=x.dtype)
        new_attn_mask.masked_fill_(xy_attn_mask, float("-inf"))
        xy_attn_mask = new_attn_mask
        # x 和完整的 y 一次性输入模型
        xy_pos = torch.concat([x, y_pos], dim=1)

        return xy_pos, xy_attn_mask, targets

    def forward(self, x, x_lens, y, y_lens, bert_feature):
        """
        x: phoneme_ids
        y: semantic_ids
        """

        reject_y, reject_y_lens = make_reject_y(y, y_lens)

        xy_pos, xy_attn_mask, targets = self.make_input_data(x, x_lens, y, y_lens, bert_feature)

        xy_dec, _ = self.h(
            (xy_pos, None),
            mask=xy_attn_mask,
        )
        x_len = x_lens.max()
        logits = self.ar_predict_layer(xy_dec[:, x_len:])

        ###### DPO #############
        reject_xy_pos, reject_xy_attn_mask, reject_targets = self.make_input_data(x, x_lens, reject_y, reject_y_lens, bert_feature)

        reject_xy_dec, _ = self.h(
            (reject_xy_pos, None),
            mask=reject_xy_attn_mask,
        )
        x_len = x_lens.max()
        reject_logits = self.ar_predict_layer(reject_xy_dec[:, x_len:])

        # loss
        # from feiteng: 每次 duration 越多, 梯度更新也应该更多, 所以用 sum

        loss_1 = F.cross_entropy(logits.permute(0, 2, 1), targets, reduction="sum")
        acc = self.ar_accuracy_metric(logits.permute(0, 2, 1).detach(), targets).item()

        A_logits, R_logits = get_batch_logps(logits, reject_logits, targets, reject_targets)
        loss_2, _, _ = dpo_loss(A_logits, R_logits, 0, 0, 0.2, reference_free=True)
        
        loss = loss_1 + loss_2

        return loss, acc

    def forward_old(self, x, x_lens, y, y_lens, bert_feature):
        """
        x: phoneme_ids
        y: semantic_ids
        """
        x = self.ar_text_embedding(x)
        x = x + self.bert_proj(bert_feature.transpose(1, 2))
        x = self.ar_text_position(x)
        x_mask = make_pad_mask(x_lens)

        y_mask = make_pad_mask(y_lens)
        y_mask_int = y_mask.type(torch.int64)
        codes = y.type(torch.int64) * (1 - y_mask_int)

        # Training
        # AR Decoder
        y, targets = self.pad_y_eos(codes, y_mask_int, eos_id=self.EOS)
        x_len = x_lens.max()
        y_len = y_lens.max()
        y_emb = self.ar_audio_embedding(y)
        y_pos = self.ar_audio_position(y_emb)

        xy_padding_mask = torch.concat([x_mask, y_mask], dim=1)
        ar_xy_padding_mask = xy_padding_mask

        x_attn_mask = F.pad(
            torch.zeros((x_len, x_len), dtype=torch.bool, device=x.device),
            (0, y_len),
            value=True,
        )
        y_attn_mask = F.pad(
            torch.triu(
                torch.ones(y_len, y_len, dtype=torch.bool, device=x.device),
                diagonal=1,
            ),
            (x_len, 0),
            value=False,
        )
        xy_attn_mask = torch.concat([x_attn_mask, y_attn_mask], dim=0)
        bsz, src_len = x.shape[0], x_len + y_len
        _xy_padding_mask = (
            ar_xy_padding_mask.view(bsz, 1, 1, src_len)
            .expand(-1, self.num_head, -1, -1)
            .reshape(bsz * self.num_head, 1, src_len)
        )
        xy_attn_mask = xy_attn_mask.logical_or(_xy_padding_mask)
        new_attn_mask = torch.zeros_like(xy_attn_mask, dtype=x.dtype)
        new_attn_mask.masked_fill_(xy_attn_mask, float("-inf"))
        xy_attn_mask = new_attn_mask
        # x 和完整的 y 一次性输入模型
        xy_pos = torch.concat([x, y_pos], dim=1)
        xy_dec, _ = self.h(
            (xy_pos, None),
            mask=xy_attn_mask,
        )
        logits = self.ar_predict_layer(xy_dec[:, x_len:]).permute(0, 2, 1)
        # loss
        # from feiteng: 每次 duration 越多, 梯度更新也应该更多, 所以用 sum
        loss = F.cross_entropy(logits, targets, reduction="sum")
        acc = self.ar_accuracy_metric(logits.detach(), targets).item()
        return loss, acc

    # 需要看下这个函数和 forward 的区别以及没有 semantic 的时候 prompts 输入什么
    def infer(
            self,
            x,
            x_lens,
            prompts,
            bert_feature,
            top_k: int = -100,
            early_stop_num: int = -1,
            temperature: float = 1.0,
    ):
        x = self.ar_text_embedding(x)
        x = x + self.bert_proj(bert_feature.transpose(1, 2))
        x = self.ar_text_position(x)

        # AR Decoder
        y = prompts
        prefix_len = y.shape[1]
        x_len = x.shape[1]
        x_attn_mask = torch.zeros((x_len, x_len), dtype=torch.bool)
        stop = False
        for _ in tqdm(range(1500)):
            y_emb = self.ar_audio_embedding(y)
            y_pos = self.ar_audio_position(y_emb)
            # x 和逐渐增长的 y 一起输入给模型
            xy_pos = torch.concat([x, y_pos], dim=1)
            y_len = y.shape[1]
            x_attn_mask_pad = F.pad(
                x_attn_mask,
                (0, y_len),
                value=True,
            )
            y_attn_mask = F.pad(
                torch.triu(torch.ones(y_len, y_len, dtype=torch.bool), diagonal=1),
                (x_len, 0),
                value=False,
            )
            xy_attn_mask = torch.concat([x_attn_mask_pad, y_attn_mask], dim=0).to(
                y.device
            )

            xy_dec, _ = self.h(
                (xy_pos, None),
                mask=xy_attn_mask,
            )
            logits = self.ar_predict_layer(xy_dec[:, -1])
            samples = topk_sampling(
                logits, top_k=top_k, top_p=1.0, temperature=temperature
            )

            if early_stop_num != -1 and (y.shape[1] - prefix_len) > early_stop_num:
                print("use early stop num:", early_stop_num)
                stop = True

            if torch.argmax(logits, dim=-1)[0] == self.EOS or samples[0, 0] == self.EOS:
                # print(torch.argmax(logits, dim=-1)[0] == self.EOS, samples[0, 0] == self.EOS)
                stop = True
            if stop:
                if prompts.shape[1] == y.shape[1]:
                    y = torch.concat([y, torch.zeros_like(samples)], dim=1)
                    print("bad zero prediction")
                print(f"T2S Decoding EOS [{prefix_len} -> {y.shape[1]}]")
                break
            # 本次生成的 semantic_ids 和之前的 y 构成新的 y
            # print(samples.shape)#[1,1]#第一个1是bs
            # import os
            # os._exit(2333)
            y = torch.concat([y, samples], dim=1)
        return y

    def pad_y_eos(self, y, y_mask_int, eos_id):
        targets = F.pad(y, (0, 1), value=0) + eos_id * F.pad(
            y_mask_int, (0, 1), value=1
        )
        # 错位
        return targets[:, :-1], targets[:, 1:]

    def infer_panel_batch_infer(
        self,
        x:List[torch.LongTensor],  #####全部文本token
        x_lens:torch.LongTensor,
        prompts:torch.LongTensor,  ####参考音频token
        bert_feature:List[torch.LongTensor],
        top_k: int = -100,
        top_p: int = 100,
        early_stop_num: int = -1,
        temperature: float = 1.0,
        repetition_penalty: float = 1.35,
        **kwargs,
    ):
        if prompts is None:
            print("Warning: Prompt free is not supported batch_infer! switch to naive_infer")
            return self.infer_panel_naive_batched(x, x_lens, prompts, bert_feature, top_k=top_k, top_p=top_p, early_stop_num=early_stop_num, temperature=temperature, **kwargs)


        max_len = kwargs.get("max_len",x_lens.max())
        x_list = []
        for x_item, bert_item in zip(x, bert_feature):
            # max_len = max(max_len, x_item.shape[0], bert_item.shape[1])
            x_item = self.ar_text_embedding(x_item.unsqueeze(0))
            x_item = x_item + self.bert_proj(bert_item.transpose(0, 1).unsqueeze(0))
            x_item = self.ar_text_position(x_item).squeeze(0)
            x_item = F.pad(x_item,(0,0,0,max_len-x_item.shape[0]),value=0) if x_item.shape[0]<max_len else x_item
            x_list.append(x_item)
        x = torch.stack(x_list, dim=0)


        # AR Decoder
        y = prompts
        
        x_len = x.shape[1]
        x_attn_mask = torch.zeros((x_len, x_len), dtype=torch.bool)
        stop = False

        k_cache = None
        v_cache = None
        ###################  first step ##########################
        if y is not None:
            y_emb = self.ar_audio_embedding(y)
            y_len = y_emb.shape[1]
            prefix_len = y.shape[1]
            y_lens = torch.LongTensor([y_emb.shape[1]]*y_emb.shape[0]).to(x.device)
            y_pos = self.ar_audio_position(y_emb)
            xy_pos = torch.concat([x, y_pos], dim=1)
            ref_free = False
        else:
            y_emb = None
            y_len = 0
            prefix_len = 0
            y_lens = torch.LongTensor([y_len]*x.shape[0]).to(x.device)
            y_pos = None
            xy_pos = x
            y = torch.zeros(x.shape[0], 0, dtype=torch.int, device=x.device)
            ref_free = True


        ##### create mask #####
        bsz = x.shape[0]
        src_len = x_len + y_len
        y_paddind_mask = make_pad_mask(y_lens, y_len)
        x_paddind_mask = make_pad_mask(x_lens, max_len)
        
        # (bsz, x_len + y_len)
        xy_padding_mask = torch.concat([x_paddind_mask, y_paddind_mask], dim=1)

        x_mask = F.pad(
            x_attn_mask,
            (0, y_len),  ###xx的纯0扩展到xx纯0+xy纯1，(x,x+y)
            value=True,
        )
        y_mask = F.pad(  ###yy的右上1扩展到左边xy的0,(y,x+y)
            torch.triu(torch.ones(y_len, y_len, dtype=torch.bool), diagonal=1), 
            (x_len, 0),
            value=False,
        )
        
        xy_mask = torch.concat([x_mask, y_mask], dim=0).view(1 , src_len, src_len).repeat(bsz, 1, 1).to(x.device)
        _xy_padding_mask = xy_padding_mask.view(bsz, 1, src_len).repeat(1, src_len, 1)
        
        for i in range(bsz):
            l = x_lens[i]
            _xy_padding_mask[i,l:max_len,:]=True
            
        xy_attn_mask = xy_mask.logical_or(_xy_padding_mask)
        xy_attn_mask = xy_attn_mask.unsqueeze(1).expand(-1, self.num_head, -1, -1)
        xy_attn_mask = xy_attn_mask.bool()
        xy_padding_mask = xy_padding_mask.view(bsz, src_len, 1).expand(-1, -1, self.model_dim)

        ###### decode #####
        y_list = [None]*y.shape[0]
        batch_idx_map = list(range(y.shape[0]))
        idx_list = [None]*y.shape[0]
        for idx in tqdm(range(1500)):
            if idx == 0:
                xy_dec, k_cache, v_cache = self.t2s_transformer.process_prompt(xy_pos, xy_attn_mask, xy_padding_mask, False)
            else:
                xy_dec, k_cache, v_cache = self.t2s_transformer.decode_next_token(xy_pos, k_cache, v_cache, xy_attn_mask, False)
            logits = self.ar_predict_layer(
                xy_dec[:, -1]
            )

            if idx == 0:
                xy_attn_mask = F.pad(xy_attn_mask[:,:,-1].unsqueeze(-2),(0,1),value=False)
                logits = logits[:, :-1]
            else:
                xy_attn_mask = F.pad(xy_attn_mask,(0,1),value=False)

            samples = sample(
                    logits, y, top_k=top_k, top_p=top_p, repetition_penalty=repetition_penalty, temperature=temperature
                )[0]

            y = torch.concat([y, samples], dim=1)
            
            ####### 移除batch中已经生成完毕的序列,进一步优化计算量
            tokens = torch.argmax(logits, dim=-1)
            reserved_idx_of_batch_for_y = None
            if (self.EOS in samples[:, 0]) or \
                (self.EOS in tokens):  ###如果生成到EOS，则停止
                    l1 = samples[:, 0]==self.EOS
                    l2 = tokens==self.EOS
                    l = l1.logical_or(l2)
                    removed_idx_of_batch_for_y = torch.where(l==True)[0].tolist()
                    reserved_idx_of_batch_for_y = torch.where(l==False)[0]
                    # batch_indexs = torch.tensor(batch_idx_map, device=y.device)[removed_idx_of_batch_for_y]
                    for i in removed_idx_of_batch_for_y:
                        batch_index = batch_idx_map[i]
                        idx_list[batch_index] = idx - 1
                        y_list[batch_index] = y[i, :-1]
                
                    batch_idx_map = [batch_idx_map[i] for i in reserved_idx_of_batch_for_y.tolist()]
                
            # 只保留batch中未生成完毕的序列 
            if reserved_idx_of_batch_for_y is not None:
                # index = torch.LongTensor(batch_idx_map).to(y.device)
                y = torch.index_select(y, dim=0, index=reserved_idx_of_batch_for_y)
                xy_attn_mask = torch.index_select(xy_attn_mask, dim=0, index=reserved_idx_of_batch_for_y)
                if k_cache is not None :
                    for i in range(len(k_cache)):
                        k_cache[i] = torch.index_select(k_cache[i], dim=0, index=reserved_idx_of_batch_for_y)
                        v_cache[i] = torch.index_select(v_cache[i], dim=0, index=reserved_idx_of_batch_for_y)
                
                
            if (early_stop_num != -1 and (y.shape[1] - prefix_len) > early_stop_num) or idx==1499:
                print("use early stop num:", early_stop_num)
                stop = True
                for i, batch_index in enumerate(batch_idx_map):
                    batch_index = batch_idx_map[i]
                    idx_list[batch_index] = idx
                    y_list[batch_index] = y[i, :-1]
                
            if not (None in idx_list):
                stop = True
                
            if stop:
                if y.shape[1]==0:
                    y = torch.concat([y, torch.zeros_like(samples)], dim=1)
                    print("bad zero prediction")
                print(f"T2S Decoding EOS [{prefix_len} -> {y.shape[1]}]")
                break

            ####################### update next step ###################################
            y_emb = self.ar_audio_embedding(y[:, -1:])
            xy_pos = y_emb * self.ar_audio_position.x_scale + self.ar_audio_position.alpha * self.ar_audio_position.pe[:, y_len + idx].to( dtype= y_emb.dtype,device=y_emb.device)            

        if (None in idx_list):
            for i in range(x.shape[0]):
                if idx_list[i] is None:
                    idx_list[i] = 1500-1  ###如果没有生成到EOS，就用最大长度代替
                    
        if ref_free:
            return y_list, [0]*x.shape[0]
        # print(idx_list)
        return y_list, idx_list
    
    def infer_panel_naive_batched(self,
        x:List[torch.LongTensor],  #####全部文本token
        x_lens:torch.LongTensor,
        prompts:torch.LongTensor,  ####参考音频token
        bert_feature:List[torch.LongTensor],
        top_k: int = -100,
        top_p: int = 100,
        early_stop_num: int = -1,
        temperature: float = 1.0,
        repetition_penalty: float = 1.35,
        **kwargs
        ):
        y_list = []
        idx_list = []
        for i in range(len(x)):
            y, idx = self.infer_panel_naive(x[i].unsqueeze(0), 
                                                  x_lens[i], 
                                                  prompts[i].unsqueeze(0) if prompts is not None else None, 
                                                  bert_feature[i].unsqueeze(0), 
                                                  top_k, 
                                                  top_p, 
                                                  early_stop_num, 
                                                  temperature,
                                                  repetition_penalty,
                                                  **kwargs)
            y_list.append(y[0])
            idx_list.append(idx)
        
        return y_list, idx_list
    
    def infer_panel_naive(
        self,
        x:torch.LongTensor,  #####全部文本token
        x_lens:torch.LongTensor,
        prompts:torch.LongTensor,  ####参考音频token
        bert_feature:torch.LongTensor,
        top_k: int = -100,
        top_p: int = 100,
        early_stop_num: int = -1,
        temperature: float = 1.0,
        repetition_penalty: float = 1.35,
        **kwargs
    ):
        x = self.ar_text_embedding(x)
        x = x + self.bert_proj(bert_feature.transpose(1, 2))
        x = self.ar_text_position(x)

        # AR Decoder
        y = prompts

        x_len = x.shape[1]
        x_attn_mask = torch.zeros((x_len, x_len), dtype=torch.bool)
        stop = False
        # print(1111111,self.num_layers)

        k_cache = None
        v_cache = None
        ###################  first step ##########################
        if y is not None:
            y_emb = self.ar_audio_embedding(y)
            y_len = y_emb.shape[1]
            prefix_len = y.shape[1]
            y_pos = self.ar_audio_position(y_emb)
            xy_pos = torch.concat([x, y_pos], dim=1)
            ref_free = False
        else:
            y_emb = None
            y_len = 0
            prefix_len = 0
            y_pos = None
            xy_pos = x
            y = torch.zeros(x.shape[0], 0, dtype=torch.int, device=x.device)
            ref_free = True

        bsz = x.shape[0]
        src_len = x_len + y_len
        x_attn_mask_pad = F.pad(
            x_attn_mask,
            (0, y_len),  ###xx的纯0扩展到xx纯0+xy纯1，(x,x+y)
            value=True,
        )
        y_attn_mask = F.pad(  ###yy的右上1扩展到左边xy的0,(y,x+y)
            torch.triu(torch.ones(y_len, y_len, dtype=torch.bool), diagonal=1),
            (x_len, 0),
            value=False,
        )
        xy_attn_mask = torch.concat([x_attn_mask_pad, y_attn_mask], dim=0)\
                                                .unsqueeze(0)\
                                                .expand(bsz*self.num_head, -1, -1)\
                                                .view(bsz, self.num_head, src_len, src_len)\
                                                .to(device=x.device, dtype=torch.bool)

        for idx in tqdm(range(1500)):
            if xy_attn_mask is not None:
                xy_dec, k_cache, v_cache = self.t2s_transformer.process_prompt(xy_pos, xy_attn_mask, None)
            else:
                xy_dec, k_cache, v_cache = self.t2s_transformer.decode_next_token(xy_pos, k_cache, v_cache)

            logits = self.ar_predict_layer(
                xy_dec[:, -1]
            )

            if idx == 0:
                xy_attn_mask = None
            if(idx<11):###至少预测出10个token不然不给停止（0.4s）
                logits = logits[:, :-1]

            samples = sample(
                logits, y, top_k=top_k, top_p=top_p, repetition_penalty=repetition_penalty, temperature=temperature
            )[0]

            y = torch.concat([y, samples], dim=1)

            if early_stop_num != -1 and (y.shape[1] - prefix_len) > early_stop_num:
                print("use early stop num:", early_stop_num)
                stop = True

            if torch.argmax(logits, dim=-1)[0] == self.EOS or samples[0, 0] == self.EOS:
                stop = True
            if stop:
                if y.shape[1] == 0:
                    y = torch.concat([y, torch.zeros_like(samples)], dim=1)
                    print("bad zero prediction")
                print(f"T2S Decoding EOS [{prefix_len} -> {y.shape[1]}]")
                break

            ####################### update next step ###################################
            y_emb = self.ar_audio_embedding(y[:, -1:])
            xy_pos = y_emb * self.ar_audio_position.x_scale + self.ar_audio_position.alpha * self.ar_audio_position.pe[:, y_len + idx].to(dtype=y_emb.dtype,device=y_emb.device)

        if ref_free:
            return y[:, :-1], 0
        return y[:, :-1], idx - 1
    
    
    def infer_panel(
        self,
        x:torch.LongTensor,  #####全部文本token
        x_lens:torch.LongTensor,
        prompts:torch.LongTensor,  ####参考音频token
        bert_feature:torch.LongTensor,
        top_k: int = -100,
        top_p: int = 100,
        early_stop_num: int = -1,
        temperature: float = 1.0,
        repetition_penalty: float = 1.35,
        **kwargs
    ):
        return self.infer_panel_naive(x, x_lens, prompts, bert_feature, top_k, top_p, early_stop_num, temperature, repetition_penalty, **kwargs)
